from typing import Callable, Optional, Union
import functools
import math
import os

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn

from realhf.impl.model.parallelism.model_parallel.modules import (
    ColumnParallelLinear,
    merged_linear_with_grad_accumulation_and_async_allreduce,
    RowParallelLinear,
)
import realhf.base.constants as constants
import realhf.base.logging as logging

logger = logging.getLogger("Modules")


def get_activation_fn(activation_function: str) -> Callable:
    if activation_function == "gelu":
        return nn.functional.gelu
    elif activation_function == "gelu_new":
        from .activations import new_gelu_activation

        return new_gelu_activation
    elif activation_function == "silu":
        return nn.SiLU()
    else:
        raise NotImplementedError(
            'Only "gelu" activation function is available.'
        )


SEQUENCE_PARALLEL_WARNED = False


class LayerNormQKVLinear(nn.Module):

    def __init__(
        self,
        input_dim: int,
        head_dim: int,
        n_q_heads: int,
        n_kv_heads: int,
        layer_norm_epsilon: float,
        use_attention_bias: bool,
        layer_norm_type: Optional[str] = None,
        # parallelism
        model_parallel: bool = False,  # We set this as an option for replacing this module with layers in transformer engine
        gradient_accumulation_fusion: bool = False,
        # dtype and device
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        layer_index=None,
    ):
        super().__init__()
        sequence_parallel = constants.sequence_parallel()
        if not model_parallel and (
            sequence_parallel or gradient_accumulation_fusion
        ):
            global SEQUENCE_PARALLEL_WARNED
            if not SEQUENCE_PARALLEL_WARNED:
                logger.warning(
                    "sequence_parallel and gradient_accumulation_fusion are only available in model parallel mode"
                )
                SEQUENCE_PARALLEL_WARNED = True
            sequence_parallel = False
            gradient_accumulation_fusion = False
        if dtype is None:
            dtype = torch.float16
        if layer_norm_type is None:
            layer_norm_fn = nn.LayerNorm
        elif layer_norm_type == "rms":
            layer_norm_fn = LlamaRMSNorm
        self.ln = layer_norm_fn(
            input_dim, eps=layer_norm_epsilon, dtype=dtype, device=device
        )

        self.model_parallel = model_parallel
        self.layer_index = layer_index
        if not model_parallel:
            self.q_attn = nn.Linear(
                input_dim,
                head_dim * n_q_heads,
                bias=use_attention_bias,
                dtype=dtype,
                device=device,
            )
            self.k_attn = nn.Linear(
                input_dim,
                head_dim * n_kv_heads,
                bias=use_attention_bias,
                dtype=dtype,
                device=device,
            )
            self.v_attn = nn.Linear(
                input_dim,
                head_dim * n_kv_heads,
                bias=use_attention_bias,
                dtype=dtype,
                device=device,
            )
        else:
            self.mp_worldsize = constants.model_parallel_world_size()
            assert n_q_heads % self.mp_worldsize == 0, (
                f"n_q_heads {n_q_heads} must be divisible by "
                f"mp_worldsize {self.mp_worldsize}"
            )
            hidden_dim = input_dim
            self.q_attn = ColumnParallelLinear(
                hidden_dim,
                head_dim * n_q_heads,
                bias=use_attention_bias,
                async_tensor_model_parallel_allreduce=not sequence_parallel,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
            if n_kv_heads > 1 and n_kv_heads % self.mp_worldsize == 0:
                # split model parallel among heads if possible
                self.k_attn = ColumnParallelLinear(
                    hidden_dim,
                    head_dim * n_kv_heads,
                    bias=use_attention_bias,
                    async_tensor_model_parallel_allreduce=not sequence_parallel,
                    gradient_accumulation_fusion=gradient_accumulation_fusion,
                    dtype=dtype,
                    device=device,
                )
                self.v_attn = ColumnParallelLinear(
                    hidden_dim,
                    head_dim * n_kv_heads,
                    bias=use_attention_bias,
                    async_tensor_model_parallel_allreduce=not sequence_parallel,
                    gradient_accumulation_fusion=gradient_accumulation_fusion,
                    dtype=dtype,
                    device=device,
                )
            else:
                if n_kv_heads > 1:
                    logger.warning(
                        f"Cannot split {n_kv_heads} kv heads evenly among "
                        f"{self.mp_worldsize} model parallel ranks, "
                        f"use unsplitted linear for kv heads instead"
                    )
                self.k_attn = nn.Linear(
                    hidden_dim,
                    head_dim * n_kv_heads,
                    bias=use_attention_bias,
                    dtype=dtype,
                    device=device,
                )
                self.v_attn = nn.Linear(
                    hidden_dim,
                    head_dim * n_kv_heads,
                    bias=use_attention_bias,
                    dtype=dtype,
                    device=device,
                )
                dist.all_reduce(
                    self.k_attn.weight.data,
                    op=dist.ReduceOp.SUM,
                    group=constants.model_parallel_group(),
                )
                if use_attention_bias:
                    dist.all_reduce(
                        self.k_attn.bias.data,
                        op=dist.ReduceOp.SUM,
                        group=constants.model_parallel_group(),
                    )
                dist.all_reduce(
                    self.v_attn.weight.data,
                    op=dist.ReduceOp.SUM,
                    group=constants.model_parallel_group(),
                )
                if use_attention_bias:
                    dist.all_reduce(
                        self.v_attn.bias.data,
                        op=dist.ReduceOp.SUM,
                        group=constants.model_parallel_group(),
                    )

        self.d = head_dim
        self.nq = n_q_heads
        self.nkv = n_kv_heads

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        hidden_states = self.ln(x)
        if not self.model_parallel:
            q = self.q_attn(hidden_states)
            k = self.k_attn(hidden_states)
            v = self.v_attn(hidden_states)
            # qkv = self.linear(hidden_states)
            # q, k, v = torch.split(qkv, (self.d * self.nq, self.d * self.nkv, self.d * self.nkv), dim=-1)
            q = q.view(*q.shape[:-1], self.nq, self.d)
            k = k.view(*k.shape[:-1], self.nkv, self.d)
            v = v.view(*v.shape[:-1], self.nkv, self.d)
        else:
            _gradient_accumulation_fusion = (
                self.q_attn.gradient_accumulation_fusion
            )
            _async_grad_allreduce = (
                self.q_attn.async_tensor_model_parallel_allreduce
            )
            _sequence_parallel = constants.sequence_parallel()
            _is_w_parallel = [
                True,
                isinstance(self.k_attn, ColumnParallelLinear),
                isinstance(self.v_attn, ColumnParallelLinear),
            ]
            q, k, v = merged_linear_with_grad_accumulation_and_async_allreduce(
                hidden_states,
                _gradient_accumulation_fusion,
                _async_grad_allreduce,
                _sequence_parallel,
                _is_w_parallel,
                self.q_attn.weight,
                self.q_attn.bias,
                self.k_attn.weight,
                self.k_attn.bias,
                self.v_attn.weight,
                self.v_attn.bias,
            )
            q = q.view(*q.shape[:-1], self.nq // self.mp_worldsize, self.d)
            if isinstance(self.k_attn, ColumnParallelLinear):
                k = k.view(*k.shape[:-1], self.nkv // self.mp_worldsize, self.d)
                v = v.view(*v.shape[:-1], self.nkv // self.mp_worldsize, self.d)
            else:
                k = k.view(*k.shape[:-1], self.nkv, self.d)
                v = v.view(*v.shape[:-1], self.nkv, self.d)
        return q, k, v


class LayerNormMLP(nn.Module):
    SEQUENCE_PARALLEL_WARNED = False

    def __init__(
        self,
        hidden_dim: int,
        intermediate_dim: int,
        resid_pdrop: float,
        activation_function: str,
        layer_norm_epsilon: float,
        # parallelism
        model_parallel: bool = False,  # We set this as an option for replacing this module with layers in transformer engine
        gradient_accumulation_fusion: bool = False,
        # dtype and device
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        sequence_parallel = constants.sequence_parallel()
        if not model_parallel and (
            sequence_parallel or gradient_accumulation_fusion
        ):
            global SEQUENCE_PARALLEL_WARNED
            if not SEQUENCE_PARALLEL_WARNED:
                logger.warning(
                    "sequence_parallel and gradient_accumulation_fusion are only available in model parallel mode"
                )
                SEQUENCE_PARALLEL_WARNED = True
            sequence_parallel = False
            gradient_accumulation_fusion = False
        if dtype is None:
            dtype = torch.float16

        self.ln = nn.LayerNorm(
            hidden_dim, eps=layer_norm_epsilon, dtype=dtype, device=device
        )
        self.model_parallel = model_parallel
        if not model_parallel:
            self.c_fc = nn.Linear(
                hidden_dim, intermediate_dim, dtype=dtype, device=device
            )
            self.c_proj = nn.Linear(
                intermediate_dim, hidden_dim, dtype=dtype, device=device
            )
        else:
            self.c_fc = ColumnParallelLinear(
                hidden_dim,
                intermediate_dim,
                async_tensor_model_parallel_allreduce=not sequence_parallel,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
            self.c_proj = RowParallelLinear(
                intermediate_dim,
                hidden_dim,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
        self.act = get_activation_fn(activation_function)
        self.dropout = nn.Dropout(resid_pdrop)

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        hidden_states = self.ln(hidden_states)
        hidden_states = self.c_fc(hidden_states)
        hidden_states = self.act(hidden_states)
        hidden_states = self.c_proj(hidden_states)
        return self.dropout(hidden_states)


class LlamaLayerNormMLP(nn.Module):

    def __init__(
        self,
        hidden_dim: int,
        intermediate_dim: int,
        activation_function: str,
        layer_norm_epsilon: float,
        # parallelism
        model_parallel: bool = False,  # We set this as an option for replacing this module with layers in transformer engine
        gradient_accumulation_fusion: bool = False,
        # dtype and device
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        sequence_parallel = constants.sequence_parallel()
        if not model_parallel and (
            sequence_parallel or gradient_accumulation_fusion
        ):
            global SEQUENCE_PARALLEL_WARNED
            if not SEQUENCE_PARALLEL_WARNED:
                logger.warning(
                    "sequence_parallel and gradient_accumulation_fusion are only available in model parallel mode"
                )
                SEQUENCE_PARALLEL_WARNED = True
            sequence_parallel = False
            gradient_accumulation_fusion = False

        if dtype is None:
            dtype = torch.float16
        self.hidden_size = hidden_dim
        self.intermediate_size = intermediate_dim
        self.ln = LlamaRMSNorm(
            hidden_dim, eps=layer_norm_epsilon, dtype=dtype, device=device
        )

        self.model_parallel = model_parallel
        if not model_parallel:
            self.gate_proj = nn.Linear(
                self.hidden_size,
                self.intermediate_size,
                bias=False,
                dtype=dtype,
                device=device,
            )
            self.up_proj = nn.Linear(
                self.hidden_size,
                self.intermediate_size,
                bias=False,
                dtype=dtype,
                device=device,
            )
            self.down_proj = nn.Linear(
                self.intermediate_size,
                self.hidden_size,
                bias=False,
                dtype=dtype,
                device=device,
            )
        else:
            self.gate_proj = ColumnParallelLinear(
                self.hidden_size,
                self.intermediate_size,
                async_tensor_model_parallel_allreduce=not sequence_parallel,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                bias=False,
                dtype=dtype,
                device=device,
            )
            self.up_proj = ColumnParallelLinear(
                self.hidden_size,
                self.intermediate_size,
                async_tensor_model_parallel_allreduce=not sequence_parallel,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                bias=False,
                dtype=dtype,
                device=device,
            )
            self.down_proj = RowParallelLinear(
                self.intermediate_size,
                self.hidden_size,
                gradient_accumulation_fusion=gradient_accumulation_fusion,
                bias=False,
                dtype=dtype,
                device=device,
            )
        self.act_fn = get_activation_fn(activation_function)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.ln(x)
        if not self.model_parallel:
            return self.down_proj(
                self.act_fn(self.gate_proj(x)) * self.up_proj(x)
            )
        else:
            _gradient_accumulation_fusion = (
                self.gate_proj.gradient_accumulation_fusion
            )
            _async_grad_allreduce = (
                self.gate_proj.async_tensor_model_parallel_allreduce
            )
            _sequence_parallel = constants.sequence_parallel()
            _is_w_parallel = [True, True]
            gate, upproj = (
                merged_linear_with_grad_accumulation_and_async_allreduce(
                    x,
                    _gradient_accumulation_fusion,
                    _async_grad_allreduce,
                    _sequence_parallel,
                    _is_w_parallel,
                    self.gate_proj.weight,
                    self.gate_proj.bias,
                    self.up_proj.weight,
                    self.up_proj.bias,
                )
            )
            return self.down_proj(self.act_fn(gate) * upproj)


class _LlamaRMSNorm(nn.Module):

    def __init__(
        self,
        hidden_size: int,
        eps: float = 1e-6,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        """
        LlamaRMSNorm is equivalent to T5LayerNorm
        """
        super().__init__()
        self.weight = nn.Parameter(
            torch.ones(hidden_size, dtype=dtype, device=device)
        )
        self.variance_epsilon = eps

    def forward(self, hidden_states: torch.Tensor):
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.to(torch.float32)
        variance = hidden_states.pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(
            variance + self.variance_epsilon
        )
        return self.weight * hidden_states.to(input_dtype)


if constants.use_te_impl():
    try:
        # HACK: we use transformer engine's rms norm as long as we can find the transformer engine package
        import transformer_engine.pytorch as te

        def _TELlamaRMSNorm(
            hidden_size: int,
            eps: float = 1e-6,
            dtype: Optional[torch.dtype] = None,
            device: Optional[Union[str, torch.device]] = None,
        ):
            return te.module.rmsnorm.RMSNorm(
                hidden_size=hidden_size,
                eps=eps,
                sequence_parallel=constants.sequence_parallel(),
                params_dtype=dtype,
                device=device,
            )

        LlamaRMSNorm = _TELlamaRMSNorm
    except ModuleNotFoundError:
        LlamaRMSNorm = _LlamaRMSNorm
    except ImportError:
        LlamaRMSNorm = _LlamaRMSNorm
else:
    LlamaRMSNorm = _LlamaRMSNorm

if constants.use_te_impl():
    from transformer_engine.pytorch.module.layernorm_mlp import (
        LayerNormMLP as _TELayerNormMLP,
    )

    # The same signature as LlamaLayerNormMLP
    def LlamaLayerNormMLP(
        hidden_dim: int,
        intermediate_dim: int,
        activation_function: str,
        layer_norm_epsilon: float,
        # parallelism
        model_parallel: bool = False,  # We set this as an option for replacing this module with layers in transformer engine
        gradient_accumulation_fusion: bool = False,
        # dtype and device
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        assert activation_function == "silu"
        return _TELayerNormMLP(
            hidden_size=hidden_dim,
            ffn_hidden_size=intermediate_dim,
            eps=layer_norm_epsilon,
            sequence_parallel=constants.sequence_parallel(),
            return_bias=False,
            tp_group=constants.model_parallel_group(),
            tp_size=constants.model_parallel_world_size(),
            bias=False,
            normalization="RMSNorm",
            activation="swiglu",
            fuse_wgrad_accumulation=gradient_accumulation_fusion,
            params_dtype=dtype,
            set_parallel_mode=model_parallel,
            device=device,
        )



from typing import Callable, Optional, Union
import functools
import math

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn

from realhf.base.constants import data_parallel_group
import realhf.base.logging as logging

logger = logging.getLogger("Modules")


class ExponentialRunningMeanStd(nn.Module):

    def __init__(self, beta=0.999, epsilon=1e-5, high_precision=True):
        super().__init__()
        self.__beta = beta
        self.__eps = epsilon

        self.__dtype = torch.float64 if high_precision else torch.float32

        self.__mean = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )
        self.__mean_sq = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )
        self.__debiasing_term = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )

        self.reset_parameters()

    def reset_parameters(self):
        self.__mean.zero_()
        self.__mean_sq.zero_()
        self.__debiasing_term.zero_()

    def forward(self, *args, **kwargs):
        # we don't implement the forward function because its meaning
        # is somewhat ambiguous
        raise NotImplementedError()

    @torch.no_grad()
    def update(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None):
        x = x.to(self.__dtype)
        if mask is not None:
            mask = mask.to(self.__dtype)
        if mask is None:
            factor = torch.tensor(
                np.prod(x.shape), dtype=self.__dtype, device=x.device
            )
        else:
            x = x * mask
            factor = mask.sum()

        x_sum = x.sum()
        x_sum_sq = x.square().sum()
        if dist.is_initialized():
            dist.all_reduce(
                factor, op=dist.ReduceOp.SUM, group=data_parallel_group()
            )
            dist.all_reduce(
                x_sum, op=dist.ReduceOp.SUM, group=data_parallel_group()
            )
            dist.all_reduce(
                x_sum_sq, op=dist.ReduceOp.SUM, group=data_parallel_group()
            )
        batch_mean = x_sum / factor
        batch_sq_mean = x_sum_sq / factor

        self.__mean.data[:] = self.__beta * self.__mean.data[:] + batch_mean * (
            1.0 - self.__beta
        )
        self.__mean_sq.data[:] = self.__beta * self.__mean_sq.data[
            :
        ] + batch_sq_mean * (1.0 - self.__beta)
        self.__debiasing_term.data[:] = (
            self.__beta * self.__debiasing_term.data[:] + 1.0 - self.__beta
        )

    @torch.no_grad()
    def mean_std(self):
        debiased_mean = self.__mean / self.__debiasing_term.clamp(
            min=self.__eps
        )
        debiased_mean_sq = self.__mean_sq / self.__debiasing_term.clamp(
            min=self.__eps
        )
        debiased_var = (debiased_mean_sq - debiased_mean**2).clamp(min=1e-2)
        return debiased_mean, debiased_var.sqrt()

    @torch.no_grad()
    def normalize(self, x):
        x = x.to(self.__dtype)
        mean, std = self.mean_std()
        return (
            ((x - mean) / std).clip(-5, 5).float()
        )  # clipping is a trick from hide and seek

    @torch.no_grad()
    def denormalize(self, x):
        x = x.to(self.__dtype)
        mean, std = self.mean_std()
        return (x * std + mean).float()


class MovingAverageRunningMeanStd(nn.Module):

    def __init__(self, high_precision=True):
        super().__init__()

        self.__dtype = torch.float64 if high_precision else torch.float32

        self.__mean = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )
        self.__mean_sq = nn.Parameter(
            torch.zeros((1,), dtype=self.__dtype, device="cuda"),
            requires_grad=False,
        )
        self.__accum_denominator = 0

        self.reset_parameters()

    def reset_parameters(self):
        self.__mean.zero_()
        self.__mean_sq.zero_()
        self.__accum_denominator = 0

    def forward(self, *args, **kwargs):
        # we don't implement the forward function because its meaning
        # is somewhat ambiguous
        raise NotImplementedError()

    @torch.no_grad()
    def update(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None):
        x = x.to(self.__dtype)
        if mask is not None:
            mask = mask.to(self.__dtype)
        if mask is None:
            factor = torch.tensor(
                np.prod(x.shape), dtype=self.__dtype, device=x.device
            )
        else:
            x = x * mask
            factor = mask.sum()

        x_sum = x.sum()
        x_sum_sq = x.square().sum()
        if dist.is_initialized():
            dist.all_reduce(
                factor, op=dist.ReduceOp.SUM, group=data_parallel_group()
            )
            dist.all_reduce(
                x_sum, op=dist.ReduceOp.SUM, group=data_parallel_group()
            )
            dist.all_reduce(
                x_sum_sq, op=dist.ReduceOp.SUM, group=data_parallel_group()
            )

        self.__mean.data[:] = (
            self.__accum_denominator * self.__mean.data[:] + x_sum
        ) / (self.__accum_denominator + factor)
        self.__mean_sq.data[:] = (
            self.__accum_denominator * self.__mean_sq.data[:] + x_sum_sq
        ) / (self.__accum_denominator + factor)
        self.__accum_denominator += factor

    @torch.no_grad()
    def mean_std(self):
        return (
            self.__mean.clone(),
            (self.__mean_sq - self.__mean**2).clamp(min=1e-2).sqrt(),
        )

    @torch.no_grad()
    def normalize(self, x):
        x = x.to(self.__dtype)
        mean, std = self.mean_std()
        return (
            ((x - mean) / std).clip(-5, 5).float()
        )  # clipping is a trick from hide and seek

    @torch.no_grad()
    def denormalize(self, x):
        x = x.to(self.__dtype)
        mean, std = self.mean_std()
        return (x * std + mean).float()


class PopArtValueHead(nn.Module):

    def __init__(
        self,
        input_dim,
        critic_dim,
        beta=0.99999,
        epsilon=1e-5,
        burn_in_updates=torch.inf,
        rms_type: str = "exp",
        high_precision=True,
    ):
        super().__init__()
        if rms_type == "exp":
            rms_cls = functools.partial(ExponentialRunningMeanStd, beta=beta)
        elif rms_type == "ma":
            rms_cls = MovingAverageRunningMeanStd
        else:
            raise NotImplementedError(f"Unknown rms type {rms_type}")
        self.__rms = rms_cls(
            input_shape=(critic_dim,),
            epsilon=epsilon,
            high_precision=high_precision,
        )

        self.__weight = nn.Parameter(torch.zeros(critic_dim, input_dim))
        self.__bias = nn.Parameter(torch.zeros(critic_dim))
        # The same initialization as `nn.Linear`.
        torch.nn.init.kaiming_uniform_(self.__weight, a=math.sqrt(5))
        torch.nn.init.uniform_(
            self.__bias, -1 / math.sqrt(input_dim), 1 / math.sqrt(input_dim)
        )

        self.__burn_in_updates = burn_in_updates
        self.__update_cnt = 0

    @property
    def weight(self):
        return self.__weight

    @property
    def bias(self):
        return self.__bias

    def forward(self, feature):
        return torch.nn.functional.linear(feature, self.__weight, self.__bias)

    @torch.no_grad()
    def update(self, x, mask):
        old_mean, old_std = self.__rms.mean_std()
        self.__rms.update(x, mask)
        new_mean, new_std = self.__rms.mean_std()
        self.__update_cnt += 1

        if self.__update_cnt > self.__burn_in_updates:
            self.__weight.data[:] = self.__weight * (
                old_std / new_std
            ).unsqueeze(-1)
            self.__bias.data[:] = (
                old_std * self.__bias + old_mean - new_mean
            ) / new_std

    @torch.no_grad()
    def normalize(self, x):
        return self.__rms.normalize(x)

    @torch.no_grad()
    def denormalize(self, x):
        return self.__rms.denormalize(x)



from typing import *
import dataclasses

from deepspeed.runtime.engine import DeepSpeedEngine, MEMORY_OPT_ALLREDUCE_SIZE
from deepspeed.runtime.zero.config import ZeroStageEnum

try:
    from megatron.core.distributed.distributed_data_parallel import (
        DistributedDataParallel as MegatronDDP,
    )
    from megatron.core.distributed.finalize_model_grads import (
        finalize_model_grads,
    )
    from megatron.core.optimizer.distrib_optimizer import (
        DistributedOptimizer as MegatronDistOptim,
    )
except ImportError or ModuleNotFoundError:
    pass
import torch
import torch.distributed as dist
import transformers

from realhf.api.core import data_api
from realhf.base.monitor import cuda_tmark, cuda_tmarked, CUDATimeMarkType
from realhf.base.namedarray import NamedArray
from realhf.impl.model.backend.utils import MegatronEngine
from realhf.impl.model.nn.real_llm_api import ReaLModel
from realhf.impl.model.nn.real_llm_base import PipeCacheData, PipeTransferData
from realhf.impl.model.nn.real_llm_generate import (
    _gather_gen_output_from_list,
    _gather_minibatch_gen_outputs,
    GenerationConfig,
    genstep,
    init_kv_cache,
)
from realhf.impl.model.parallelism.pipeline_parallel.instruction import (
    PipeInstruction,
)
from realhf.impl.model.parallelism.pipeline_parallel.static_schedule import (
    PipeSchedule,
)
from realhf.impl.model.parallelism.pipeline_parallel.tensor_storage import (
    TensorBuffer,
)
from realhf.impl.model.utils.padding import pad_sequence_parallel_input
import realhf.base.constants as constants
import realhf.base.logging as logging
import realhf.impl.model.parallelism.pipeline_parallel.p2p as p2p
import realhf.impl.model.parallelism.pipeline_parallel.static_schedule as schedule

logger = logging.getLogger("Pipeline Runner", "benchmark")


class PipelineError(Exception):
    pass


# FIXME: remove seqlens_cpu here
def _prepare_input(
    module: ReaLModel,
    tensor_buffer: TensorBuffer,
    num_micro_batches: int,
    seqlens_cpu: List[int],
    packed_input_ids: torch.Tensor,
    cu_seqlens: torch.Tensor,
    store_kv_cache: bool,
    loss_fn: Optional[Callable] = None,
    input_lens_for_partition: Optional[torch.Tensor] = None,
    **loss_kwargs,
):
    """Prepare input for train or inference
    split all input tensors into micro batches for pipeline parallel

    Args:
        packed_input_ids (torch.Tensor): packed input ids of shape [total_seq_len]
        cu_seqlens (torch.Tensor): cu_seqlens of shape [batch_size]
    """
    if input_lens_for_partition is not None:
        n_groups = input_lens_for_partition.shape[0]
        group_input_lens = (cu_seqlens[1:] - cu_seqlens[:-1]).view(n_groups, -1)
        data = NamedArray(
            packed_input_ids=packed_input_ids,
            group_input_lens=group_input_lens,
            input_lens=input_lens_for_partition,
        )
        n_seqs = input_lens_for_partition.shape[0]
    else:
        data = NamedArray(
            packed_input_ids=packed_input_ids, cu_seqlens=cu_seqlens
        )
        n_seqs = cu_seqlens.shape[0] - 1
    data.register_metadata(seqlens=seqlens_cpu)

    # split into sequences
    n_mbs = num_micro_batches
    assert n_seqs >= n_mbs
    splitted, partitions = data_api.split_sequences(
        data, n_mbs, min_size=n_seqs // n_mbs, return_partitions=True
    )
    batch_seqlens = [seqlens_cpu[start:end] for start, end in partitions]

    # Sanity check to ensure that the order of splitted sequences
    # is the same across pipeline parallel ranks.
    _batch_seqlen = torch.tensor(
        [sum(x) for x in batch_seqlens],
        device=cu_seqlens.device,
        dtype=torch.long,
    )
    _batch_seqlen_all_gathered = [
        torch.zeros_like(_batch_seqlen)
        for _ in range(constants.pipe_parallel_world_size())
    ]
    _batch_seqlen_all_gathered[constants.pipe_parallel_rank()] = _batch_seqlen
    dist.all_gather(
        _batch_seqlen_all_gathered,
        _batch_seqlen,
        group=constants.pipe_parallel_group(),
    )
    for i in range(constants.pipe_parallel_world_size()):
        if not torch.allclose(_batch_seqlen_all_gathered[i], _batch_seqlen):
            raise PipelineError(
                "Partitioned seqlens are not equal across pipeline parallel ranks. "
                f"Current rank (dp={constants.data_parallel_rank()},"
                f"tp={constants.model_parallel_rank()},pp={constants.pipe_parallel_rank()}), "
                f"gathered batch seqlens={_batch_seqlen_all_gathered}, "
                f"Have you ensured that the order of dataset across ranks is the same?",
            )

    if input_lens_for_partition is not None:
        splitted = [
            NamedArray(
                packed_input_ids=x["packed_input_ids"],
                cu_seqlens=torch.nn.functional.pad(
                    x["group_input_lens"].view(-1).cumsum(0), (1, 0)
                ),
            )
            for x in splitted
        ]

    if loss_fn is not None:
        for mbid, x in enumerate(splitted):
            tensor_buffer.put("input_cache", mbid, x)

        loss_input_data = NamedArray(**loss_kwargs)
        loss_input_data.register_metadata(seqlens=seqlens_cpu)
        splitted_loss_input = data_api.split_sequences(
            loss_input_data,
            num_micro_batches,
            min_size=n_seqs // num_micro_batches,
        )
        for mbid, x in enumerate(splitted_loss_input):
            tensor_buffer.put("loss_inputs", mbid, x)

    mb_seq_lens = []

    def input_to_pipe_model_input(input: NamedArray, mbid: int):
        max_seqlen = int(max(batch_seqlens[mbid]))

        cu_seqlens = input.cu_seqlens
        packed_input_ids = input.packed_input_ids

        # sequence parallel input padding
        if constants.sequence_parallel():
            packed_input_ids, cu_seqlens, max_seqlen, pad_size = (
                pad_sequence_parallel_input(
                    packed_input_ids, cu_seqlens, max_seqlen
                )
            )
            tensor_buffer.put("pad_size", mbid, pad_size)
        x = PipeTransferData(
            cu_seqlens=cu_seqlens.int(),
            max_seqlen=int(max_seqlen),
            store_kv_cache=store_kv_cache,
        )
        if constants.is_first_pipe_stage():
            ys = [PipeCacheData(packed_input_ids=packed_input_ids)] + [
                PipeCacheData() for _ in range(module.num_layers - 1)
            ]
        else:
            ys = [PipeCacheData() for _ in range(module.num_layers)]
        total_len = (
            packed_input_ids.shape[0]
            if not constants.sequence_parallel()
            else packed_input_ids.shape[0]
            // constants.model_parallel_world_size()
        )
        mb_seq_lens.append(total_len)
        return (x, ys)

    batches = [input_to_pipe_model_input(x, i) for i, x in enumerate(splitted)]
    for mbid, batch in enumerate(batches):
        x, ys = batch
        tensor_buffer.put("batch_input_x", mbid, x)
        tensor_buffer.put("batch_input_ys", mbid, ys)
        tensor_buffer.put("batch_lengths", mbid, x.cu_seqlens.shape[0] - 1)
        tensor_buffer.put("mb_seq_lens", mbid, mb_seq_lens[mbid])

    # pre allocate receive buffers and pre store other information
    for mbid, batch in enumerate(batches):
        others_cache = dict(
            cu_seqlens=batch[0].cu_seqlens.int(),
            max_seqlen=int(batch[0].max_seqlen),
            store_kv_cache=batch[0].store_kv_cache,
        )
        tensor_buffer.put("pipe_transfer_infos", mbid, others_cache)


def _exec_pipe_schedule(
    module: ReaLModel,
    tensor_buffer: TensorBuffer,
    instr_map: Dict[PipeInstruction, Callable],
    pipe_schedule: PipeSchedule,
    terminate_condition: Optional[Callable] = None,
):
    """Execute schedules
    Args:
        module: The model to execute the schedule on.
        tensor_buffer: A temporary buffer that stores necessary information during running.
        instr_map: A map of PipeInstruction types to methods. Each method will be executed with the
            kwargs provided to the PipeInstruction from the scheduler.
        pipe_schedule: an instance of schedule
        terminate_condition: a callable that returns boolean value indicating if
                                the pipeline execution should terminate
    """
    step_count = 0
    is_last_stage = constants.is_last_pipe_stage()
    num_stages = constants.pipe_parallel_world_size()
    stage_id = constants.pipe_parallel_rank()
    global_rank = dist.get_rank()
    parllelism_rank = constants.parallelism_rank()

    # A termination machanism to avoid all-reduce at each step.
    # If the schedule is about to terminate (i.e., will_break is True),
    # the last stage will send this message to the previous stages with
    # one more pipeline round (last -> 0 -> 1 -> .. -> last-1).
    will_break = False
    if is_last_stage:
        burn_out_steps = num_stages - 1
    elif stage_id == num_stages - 2:
        burn_out_steps = 0
    else:
        burn_out_steps = 1

    # For each step in the schedule
    for step_cmds in pipe_schedule:
        # For each instruction in the step
        step_id, micro_batch_id, step_cmds = step_cmds
        for cmd in step_cmds:
            if type(cmd) not in instr_map:
                raise RuntimeError(
                    f"Pipeline instruction executor does not understand instruction {repr(cmd)}"
                )

            if will_break:
                # With the termination mechanism, skip communication instructions
                # because its peer stages have been terminated
                if (
                    is_last_stage
                    and burn_out_steps < num_stages - 1
                    and type(cmd) != schedule.RecvActivation
                ):
                    continue
                elif not is_last_stage and type(cmd) != schedule.SendActivation:
                    continue

            try:
                instr_map[type(cmd)](module, tensor_buffer, *cmd.args)
            except Exception as e:
                logger.error(
                    f"Model name {constants.model_name()} rank {parllelism_rank}"
                    f" (global rank {global_rank}) step {step_count}, "
                    f"Exception in cmd: {cmd}"
                )
                raise e
        step_count += 1

        if will_break:
            burn_out_steps -= 1
        if terminate_condition is not None and terminate_condition():
            will_break = True
        if will_break and burn_out_steps <= 0:
            break


def _zero_grads(inputs):
    if isinstance(inputs, torch.Tensor):
        if inputs.grad is not None:
            inputs.grad.data.zero_()
    elif isinstance(inputs, tuple):
        for t in inputs:
            if t.grad is not None:
                t.grad.data.zero_()
    elif dataclasses.is_dataclass(inputs):
        for f in dataclasses.fields(inputs):
            _zero_grads(getattr(inputs, f.name))
    else:
        # do nothing for non tensor
        pass


class PipeInferenceInstrSet:

    def _exec_forward_pass(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        buf = tensor_buffer.get(
            "recv_act_buf", micro_batch_id, remove=True, raise_error=False
        )
        ys = tensor_buffer.get("batch_input_ys", micro_batch_id, remove=False)

        if buf is not None:
            others = tensor_buffer.get(
                "pipe_transfer_infos", micro_batch_id, remove=False
            )
            x = PipeTransferData(pp_input=buf, **others)
            tensor_buffer.put("batch_input_x", micro_batch_id, x)
        else:
            x = tensor_buffer.get("batch_input_x", micro_batch_id, remove=True)

        _zero_grads(x)
        _zero_grads(ys)
        x, ys = module.forward(x, ys)

        tensor_buffer.put(
            "batch_output_x", micro_batch_id, x
        )  # Used by send_activation

        if constants.is_last_pipe_stage():
            logits = x.pp_output
            tensor_buffer.put("logits", micro_batch_id, logits)

    def _exec_send_activations(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        assert stage_id != constants.pipe_parallel_world_size() - 1
        x: PipeTransferData = tensor_buffer.get(
            "batch_output_x",
            micro_batch_id,
            remove=True,
        )
        p2p.send(x.pp_output, constants.next_pipe_stage(), async_op=False)

    def _exec_recv_activations(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        assert not constants.is_first_pipe_stage()

        device = module.device
        dtype = module.dtype
        hidden_dim = module.config.hidden_dim

        mb_seq_len = tensor_buffer.get(
            "mb_seq_lens", micro_batch_id, remove=False
        )
        act_shape = (mb_seq_len, hidden_dim)
        buf = torch.empty(
            act_shape, dtype=dtype, device=device, requires_grad=False
        )

        p2p.recv(buf, constants.prev_pipe_stage(), async_op=False)
        tensor_buffer.put("recv_act_buf", micro_batch_id, buf)

    INSTRUCTION_MAP = {
        schedule.ForwardPass: _exec_forward_pass,
        schedule.SendActivation: _exec_send_activations,
        schedule.RecvActivation: _exec_recv_activations,
    }


class PipeGenInstrSet:

    def _exec_forward_pass(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        is_first_stage = constants.is_first_pipe_stage()
        if is_first_stage:
            buf = tensor_buffer.get(
                "recv_next_tokens_buf",
                micro_batch_id,
                remove=True,
                raise_error=False,
            )
        else:
            buf = tensor_buffer.get(
                "recv_act_buf",
                micro_batch_id,
                remove=True,
                raise_error=False,
            )

        ys = tensor_buffer.get("batch_input_ys", micro_batch_id, remove=False)

        if buf is not None:
            if is_first_stage:
                x = tensor_buffer.get(
                    "batch_input_x", micro_batch_id, remove=True
                )
                ys = tensor_buffer.get(
                    "batch_input_ys", micro_batch_id, remove=False
                )
                ys[0].packed_input_ids = (
                    buf  # sequence parallel forward only accept one dim input_ids
                )
                ys[0].packed_position_ids = None
            else:
                others = tensor_buffer.get(
                    "pipe_transfer_infos", micro_batch_id, remove=False
                )
                x = PipeTransferData(pp_input=buf, **others)
                tensor_buffer.put("batch_input_x", micro_batch_id, x)
        else:
            x = tensor_buffer.get("batch_input_x", micro_batch_id, remove=True)

        _zero_grads(x)
        _zero_grads(ys)
        x, ys = module.forward(x, ys)

        tensor_buffer.put("batch_output_x", micro_batch_id, x)

        tokenizer = tensor_buffer.get("tokenizer", micro_batch_id)
        gconfig = tensor_buffer.get("gconfig", micro_batch_id)

        # Init KV cache.
        is_prefill_phase = False
        if not tensor_buffer.get("kv_cache_reserved", micro_batch_id):
            # KV cache is attached to x and ys.
            assert constants.pipe_parallel_world_size() >= 2
            if constants.is_first_pipe_stage():
                ys[0].cache_seqlens = x.cu_seqlens[1:] - x.cu_seqlens[:-1]
                init_kv_cache(module, gconfig, x, ys[1:])
            elif constants.is_last_pipe_stage():
                init_kv_cache(module, gconfig, x, ys[:-1])
            else:
                init_kv_cache(module, gconfig, x, ys)
            is_prefill_phase = True
            tensor_buffer.put("kv_cache_reserved", micro_batch_id, True)

        # Increase cache_seqlens in the decoding phase.
        if not is_prefill_phase:
            if constants.is_last_pipe_stage():
                for y in ys[:-1]:
                    y.cache_seqlens += 1
                assert all(
                    torch.allclose(ys[0].cache_seqlens, y.cache_seqlens)
                    for y in ys[:-1]
                )
            else:
                for y in ys:
                    y.cache_seqlens += 1
                assert all(
                    torch.allclose(ys[0].cache_seqlens, y.cache_seqlens)
                    for y in ys
                )

        # Perform a decoding step.
        if constants.is_last_pipe_stage():
            # Gather logits of the final token
            logits = x.pp_output
            if is_prefill_phase:
                logits = logits[x.cu_seqlens[1:] - 1]

            unfinished_sequences = tensor_buffer.get(
                "unfinished_sequences", micro_batch_id
            )
            generated_idx = tensor_buffer.get("generated_idx", micro_batch_id)

            (
                next_tokens,
                logprob,
                logits_mask,
                terminate,
                unfinished_sequences,
            ) = genstep(
                logits,
                tokenizer,
                unfinished_sequences,
                generated_idx,
                gconfig,
            )

            if isinstance(terminate, bool):
                terminate = torch.tensor(
                    terminate, device=logits.device, dtype=torch.bool
                )

            tensor_buffer.put("terminate", micro_batch_id, terminate)
            tensor_buffer.put(
                "unfinished_sequences", micro_batch_id, unfinished_sequences
            )
            tensor_buffer.put(
                "generated_idx", micro_batch_id, generated_idx + 1
            )
            assert next_tokens is not None and logprob is not None
            tensor_buffer.get("gen_token_ph", micro_batch_id).append(
                next_tokens
            )
            tensor_buffer.get("gen_logprob_ph", micro_batch_id).append(logprob)
            tensor_buffer.get("gen_logits_mask_ph", micro_batch_id).append(
                logits_mask
            )
            tensor_buffer.put(
                "next_tokens_to_send", micro_batch_id, next_tokens
            )

    def _exec_send_activations(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        PipeInferenceInstrSet._exec_send_activations(
            module, tensor_buffer, stage_id, micro_batch_id, step_id
        )
        tensor_buffer.put("first_token", micro_batch_id, False)
        terminate = tensor_buffer.get("terminate", micro_batch_id)
        p2p.send(terminate, constants.next_pipe_stage())

    def _exec_recv_activations(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        assert not constants.is_first_pipe_stage()

        device = module.device
        dtype = module.dtype
        hidden_dim = module.config.hidden_dim

        mb_seq_len = tensor_buffer.get(
            "mb_seq_lens", micro_batch_id, remove=False
        )
        act_shape = (mb_seq_len, hidden_dim)

        ft = tensor_buffer.get("first_token", micro_batch_id, remove=False)
        if ft:
            buf = torch.empty(
                act_shape, dtype=dtype, device=device, requires_grad=False
            )
        else:
            batch_length = tensor_buffer.get(
                "batch_lengths", micro_batch_id, remove=False
            )
            batch_length = (
                batch_length // constants.model_parallel_world_size()
                if constants.sequence_parallel()
                else batch_length
            )
            act_shape = (batch_length, hidden_dim)
            buf = torch.empty(
                act_shape, dtype=dtype, device=device, requires_grad=False
            )

        prev_stage = constants.prev_pipe_stage()
        p2p.recv(buf, prev_stage, async_op=False)
        tensor_buffer.put("recv_act_buf", micro_batch_id, buf)

        terminate = torch.empty((), dtype=torch.bool, device=device)
        p2p.recv(terminate, prev_stage)
        tensor_buffer.put("terminate", micro_batch_id, terminate)

    def _exec_send_next_tokens(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        """When generating, send next tokens from the last stage to the first stage."""
        assert constants.is_last_pipe_stage()
        next_stage = constants.next_pipe_stage()
        next_tokens_to_send = tensor_buffer.get(
            "next_tokens_to_send", micro_batch_id, remove=True
        )
        p2p.send(next_tokens_to_send, next_stage, async_op=False)
        p2p.send(tensor_buffer.get("terminate", micro_batch_id), next_stage)
        tensor_buffer.put("first_token", micro_batch_id, False)

    def _exec_recv_next_tokens(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        """When generating, recv next tokens from the last stage on the first stage
        Construct next forward input
        """
        assert constants.is_first_pipe_stage()
        batch_length = tensor_buffer.get(
            "batch_lengths", micro_batch_id, remove=False
        )

        device = module.device
        prev_stage = constants.prev_pipe_stage()

        recv_buf = torch.empty((batch_length,), dtype=torch.long, device=device)
        p2p.recv(recv_buf, prev_stage, async_op=False)
        tensor_buffer.put("recv_next_tokens_buf", micro_batch_id, recv_buf)

        x = PipeTransferData(
            store_kv_cache=True,
            cu_seqlens=torch.arange(
                batch_length + 1, dtype=torch.int32, device=device
            ),
            max_seqlen=1,
        )
        tensor_buffer.put("batch_input_x", micro_batch_id, x)

        terminate = torch.empty((), dtype=torch.bool, device=device)
        p2p.recv(terminate, prev_stage)
        tensor_buffer.put("terminate", micro_batch_id, terminate)

    INSTRUCTION_MAP = {
        schedule.ForwardPass: _exec_forward_pass,
        schedule.SendActivation: _exec_send_activations,
        schedule.RecvActivation: _exec_recv_activations,
        schedule.SendNextTokens: _exec_send_next_tokens,
        schedule.RecvNextTokens: _exec_recv_next_tokens,
    }


class PipeTrainForwardCommInstrSet:

    def _exec_forward_pass(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        PipeInferenceInstrSet._exec_forward_pass(
            module, tensor_buffer, stage_id, micro_batch_id, step_id
        )

        loss_fn = tensor_buffer.get("loss_fn", micro_batch_id)
        if loss_fn is not None and constants.is_last_pipe_stage():
            model_output = tensor_buffer.get(
                "batch_output_x", micro_batch_id
            ).pp_output
            if constants.sequence_parallel():
                pad_size = tensor_buffer.get(
                    "pad_size", micro_batch_id, remove=True
                )
                model_output = (
                    model_output[:-pad_size] if pad_size > 0 else model_output
                )
            loss_kwargs = tensor_buffer.get(
                "loss_inputs", micro_batch_id, remove=True
            )
            input_cache = tensor_buffer.get(
                "input_cache", micro_batch_id, remove=True
            )
            packed_input_ids = input_cache.packed_input_ids
            cu_seqlens = input_cache.cu_seqlens

            loss, stats = loss_fn(
                model_output, packed_input_ids, cu_seqlens, **loss_kwargs
            )
            loss = loss / tensor_buffer.get("num_micro_batches", micro_batch_id)
            tensor_buffer.put("losses", micro_batch_id, loss)
            tensor_buffer.put("stats", micro_batch_id, stats)

    def _exec_send_activations(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        assert stage_id != constants.pipe_parallel_world_size() - 1
        # NOTE: This is different from inference, we remain batch_output_x for backward.
        x: PipeTransferData = tensor_buffer.get(
            "batch_output_x", micro_batch_id
        )
        p2p.send(x.pp_output, constants.next_pipe_stage(), async_op=False)

    def _exec_recv_activations(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        assert not constants.is_first_pipe_stage()

        device = module.device
        dtype = module.dtype
        hidden_dim = module.config.hidden_dim

        mb_seq_len = tensor_buffer.get(
            "mb_seq_lens", micro_batch_id, remove=False
        )
        act_shape = (mb_seq_len, hidden_dim)

        buf = tensor_buffer.alloc(
            "activation",
            micro_batch_id,
            act_shape,
            dtype,
            device,
            require_grads=True,
        )

        p2p.recv(buf, constants.prev_pipe_stage(), async_op=False)
        tensor_buffer.put("recv_act_buf", micro_batch_id, buf)

    def _exec_send_grads(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        assert not constants.is_first_pipe_stage()
        activation = tensor_buffer.get(
            "activation", micro_batch_id, remove=True
        )
        assert activation.grad is not None
        p2p.send(activation.grad, constants.prev_pipe_stage(), async_op=False)

    def _exec_recv_grads(
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        assert not constants.is_last_pipe_stage()
        device = module.device
        dtype = module.dtype
        hidden_dim = module.config.hidden_dim
        mb_seq_len = tensor_buffer.get(
            "mb_seq_lens", micro_batch_id, remove=False
        )
        grad_shape = (mb_seq_len, hidden_dim)
        buf = tensor_buffer.alloc(
            "grad", micro_batch_id, grad_shape, dtype, device
        )
        p2p.recv(buf, constants.next_pipe_stage(), async_op=False)

    INSTRUCTION_MAP = {
        schedule.ForwardPass: _exec_forward_pass,
        schedule.SendActivation: _exec_send_activations,
        schedule.RecvActivation: _exec_recv_activations,
        schedule.SendGrad: _exec_send_grads,
        schedule.RecvGrad: _exec_recv_grads,
    }


@dataclasses.dataclass
class PipeTrainBackwardReduceInstrSetForDeepSpeed:
    ds_engine: DeepSpeedEngine

    def __post_init__(self):
        self.ds_engine.pipeline_parallelism = True

    @cuda_tmark("bwd", CUDATimeMarkType.backward)
    def _exec_backward_pass(
        self,
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        assert self.ds_engine is not None
        assert self.ds_engine.optimizer is not None, (
            "must provide optimizer during " "init in order to use backward"
        )
        # The last stage just runs backward on the loss using DeepSpeed's typical
        # mechanisms.
        output_x = tensor_buffer.get(
            "batch_output_x", micro_batch_id, remove=True
        )

        # We schedule the all-reduces, so disable it in super().backward()
        self.ds_engine.enable_backward_allreduce = False
        self.ds_engine.set_gradient_accumulation_boundary(False)

        is_last_stage = constants.is_last_pipe_stage()
        if is_last_stage:
            loss = tensor_buffer.get("losses", micro_batch_id, remove=True)
            self.ds_engine.backward(loss)
            tensor_buffer.put("losses", micro_batch_id, loss.detach().clone())
            return False, None

        if self.ds_engine.bfloat16_enabled() and not is_last_stage:
            # manually call because we don't call optimizer.backward()
            self.ds_engine.optimizer.clear_lp_grads()

        if not is_last_stage and self.ds_engine.zero_optimization():
            # manually call because we don't call optimizer.backward()
            self.ds_engine.optimizer.micro_step_id += 1

            if self.ds_engine.optimizer.contiguous_gradients:
                self.ds_engine.optimizer.ipg_buffer = []
                buf_0 = torch.empty(
                    int(self.ds_engine.optimizer.reduce_bucket_size),
                    dtype=module.dtype,
                    device=module.device,
                )
                self.ds_engine.optimizer.ipg_buffer.append(buf_0)

                # Use double buffers to avoid data access conflict when overlap_comm is enabled.
                if self.ds_engine.optimizer.overlap_comm:
                    buf_1 = torch.empty(
                        int(self.ds_engine.optimizer.reduce_bucket_size),
                        dtype=module.dtype,
                        device=module.device,
                    )
                    self.ds_engine.optimizer.ipg_buffer.append(buf_1)
                self.ds_engine.optimizer.ipg_index = 0

        grad = tensor_buffer.get("grad", micro_batch_id, remove=True)

        output_tensor = output_x.pp_output
        torch.autograd.backward(tensors=output_tensor, grad_tensors=grad)

        if not is_last_stage and self.ds_engine.zero_optimization():
            # manually call because we don't call optimizer.backward()
            # Only for Stage 1, Mode 2
            if self.ds_engine.optimizer.use_grad_accum_attribute:
                self.ds_engine.optimizer.fill_grad_accum_attribute()

        if self.ds_engine.bfloat16_enabled() and not is_last_stage:
            # manually call because we don't call optimizer.backward()
            self.ds_engine.optimizer.update_hp_grads(clear_lp_grads=False)

    def _exec_reduce_grads(
        self,
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):

        self.ds_engine.set_gradient_accumulation_boundary(True)
        if self.ds_engine.bfloat16_enabled():
            if (
                self.ds_engine.zero_optimization_stage()
                < ZeroStageEnum.gradients
            ):
                # Make our own list of gradients from the optimizer's FP32 grads
                self.ds_engine.buffered_allreduce_fallback(
                    grads=self.ds_engine.optimizer.get_grads_for_reduction(),
                    elements_per_buffer=MEMORY_OPT_ALLREDUCE_SIZE,
                )
            else:
                raise NotImplementedError("PP+BF16 only work for ZeRO Stage 1")
        else:
            self.ds_engine.allreduce_gradients(
                bucket_size=MEMORY_OPT_ALLREDUCE_SIZE
            )

    def _exec_optimizer_step(
        self,
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        self.ds_engine.set_gradient_accumulation_boundary(True)
        version_steps = tensor_buffer.get("version_steps", 0)
        lr_kwargs = {"epoch": version_steps}
        self.ds_engine._take_model_step(lr_kwargs=lr_kwargs)

        # sync loss scale across pipeline stages
        if not self.ds_engine.bfloat16_enabled():
            loss_scale = self.ds_engine.optimizer.loss_scale
            total_scale_cuda = torch.FloatTensor([float(loss_scale)]).to(
                module.device
            )
            dist.all_reduce(
                total_scale_cuda,
                op=dist.ReduceOp.MIN,
                group=constants.grid().get_model_parallel_group(),
            )
            # all_loss_scale = total_scale_cuda[0].item()
            logger.info(
                f"loss scale: {total_scale_cuda}, "
                f"group: {dist.get_process_group_ranks(self.ds_engine.mpu.get_model_parallel_group())}"
            )
            self.ds_engine.optimizer.loss_scaler.cur_scale = min(
                total_scale_cuda[0].item(), 8192
            )

    @property
    def INSTRUCTION_MAP(self):
        return {
            schedule.OptimizerStep: self._exec_optimizer_step,
            schedule.ReduceGrads: self._exec_reduce_grads,
            schedule.BackwardPass: self._exec_backward_pass,
        }


@dataclasses.dataclass
class PipeTrainBackwardReduceInstrSetForMegatron:
    # NOTE: merge MegatronDDP and MegatronDistOptim into one class
    # to remain consistent with DeepSpeed's API
    engine: MegatronEngine
    num_micro_batches: int

    def __post_init__(self):
        self._no_sync_context = None
        self.disable_grad_sync()

    def disable_grad_sync(self):
        if self._no_sync_context is None:
            self._no_sync_context = self.engine.ddp.no_sync()
            self._no_sync_context.__enter__()

    def enable_grad_sync(self):
        if self._no_sync_context is not None:
            self._no_sync_context.__exit__(None, None, None)
            self._no_sync_context = None

    def _exec_backward_pass(
        self,
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        output_x = tensor_buffer.get(
            "batch_output_x", micro_batch_id, remove=True
        )

        if micro_batch_id == self.num_micro_batches - 1:
            self.enable_grad_sync()

        is_last_stage = constants.is_last_pipe_stage()
        if is_last_stage:
            loss: torch.Tensor = tensor_buffer.get(
                "losses", micro_batch_id, remove=True
            )
            loss = self.engine.optim.scale_loss(loss)
            loss.backward()
            tensor_buffer.put("losses", micro_batch_id, loss.detach().clone())
            return

        grad = tensor_buffer.get("grad", micro_batch_id, remove=True)
        output_tensor = output_x.pp_output
        torch.autograd.backward(tensors=output_tensor, grad_tensors=grad)

    def _exec_reduce_grads(
        self,
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        # self.engine.ddp.start_grad_sync()
        finalize_model_grads([self.engine.ddp])

    def _exec_optimizer_step(
        self,
        module: ReaLModel,
        tensor_buffer: TensorBuffer,
        stage_id: int,
        micro_batch_id: int,
        step_id: int,
    ):
        update_successful, grad_norm, num_zeros_in_grad = (
            self.engine.optim.step()
        )

        version_steps = tensor_buffer.get("version_steps", 0)
        if update_successful:
            self.engine.lr_scheduler.step_absolute(version_steps)
        if (
            constants.data_parallel_rank() == 0
            and constants.model_parallel_rank() == 0
        ):
            logger.info(
                f"Pipeline rank {constants.pipe_parallel_rank()}. "
                f"Update success? {update_successful}. "
                f"Grad Norm: {grad_norm}. "
                f"Current loss scale: {self.engine.optim.get_loss_scale()}. "
            )
        return update_successful, grad_norm, num_zeros_in_grad

    @property
    def INSTRUCTION_MAP(self):
        return {
            schedule.OptimizerStep: self._exec_optimizer_step,
            schedule.ReduceGrads: self._exec_reduce_grads,
            schedule.BackwardPass: self._exec_backward_pass,
        }


@dataclasses.dataclass
class PipeTrainInstrSet:
    backend: str = dataclasses.field(
        metadata={"choices": ["deepspeed", "megatron"]}
    )
    num_micro_batches: int
    ds_engine: Optional[DeepSpeedEngine] = None
    megatron_engine: Optional[MegatronEngine] = None

    @property
    def INSTRUCTION_MAP(self):
        if self.backend == "deepspeed":
            return {
                **PipeTrainForwardCommInstrSet.INSTRUCTION_MAP,
                **PipeTrainBackwardReduceInstrSetForDeepSpeed(
                    self.ds_engine
                ).INSTRUCTION_MAP,
            }
        elif self.backend == "megatron":
            return {
                **PipeTrainForwardCommInstrSet.INSTRUCTION_MAP,
                **PipeTrainBackwardReduceInstrSetForMegatron(
                    self.megatron_engine,
                    num_micro_batches=self.num_micro_batches,
                ).INSTRUCTION_MAP,
            }
        else:
            raise NotImplementedError()


@dataclasses.dataclass
class PipelineRunner:
    module: ReaLModel

    @property
    def default_train_mbs(self):
        return constants.pipe_parallel_world_size() * 2

    @property
    def default_inf_mbs(self):
        return constants.pipe_parallel_world_size()

    def eval(self, *args, **kwargs):
        return self.module.eval(*args, **kwargs)

    def train(self, *args, **kwargs):
        return self.module.train(*args, **kwargs)

    def forward(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        input_lens_for_partition: Optional[torch.Tensor] = None,
        num_micro_batches: Optional[int] = None,
    ):
        """Run one forward step over a batch of tokens and return the logits."""

        if num_micro_batches is None:
            num_micro_batches = self.default_inf_mbs

        tensor_buffer = TensorBuffer()

        _prepare_input(
            module=self.module,
            tensor_buffer=tensor_buffer,
            num_micro_batches=num_micro_batches,
            seqlens_cpu=seqlens_cpu,
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
            store_kv_cache=False,
            input_lens_for_partition=input_lens_for_partition,
        )

        sched = schedule.InferenceSchedule(
            micro_batches=num_micro_batches,
            stages=constants.pipe_parallel_world_size(),
            stage_id=constants.pipe_parallel_rank(),
        )
        _exec_pipe_schedule(
            self.module,
            tensor_buffer,
            instr_map=PipeInferenceInstrSet.INSTRUCTION_MAP,
            pipe_schedule=sched,
        )

        logits = None
        if constants.is_last_pipe_stage():
            logits_list = []
            for i in range(num_micro_batches):
                logits = tensor_buffer.get("logits", i, remove=True)
                # logger.info(f"mbid {i} before remove pad logits shape {logits.shape}")
                if constants.sequence_parallel():
                    pad_size = tensor_buffer.get("pad_size", i, remove=True)
                    logits = logits[:-pad_size] if pad_size > 0 else logits
                    # logger.info(f"mbid {i} after remove pad {pad_size} logits shape {logits.shape}")
                logits_list.append(logits)
            logits = torch.cat(logits_list, dim=0)

        return logits

    @torch.no_grad()
    def generate(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        tokenizer: transformers.PreTrainedTokenizerFast,
        gconfig: GenerationConfig = dataclasses.field(
            default_factory=GenerationConfig
        ),
        num_micro_batches: Optional[int] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, List[PipeCacheData]]:
        if constants.sequence_parallel():
            raise NotImplementedError(
                "Sequence parallel is not supported for generation"
            )

        if num_micro_batches is None:
            num_micro_batches = self.default_inf_mbs

        tensor_buffer = TensorBuffer()

        _prepare_input(
            module=self.module,
            tensor_buffer=tensor_buffer,
            num_micro_batches=num_micro_batches,
            seqlens_cpu=seqlens_cpu,
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
            store_kv_cache=True,
        )

        # for elegant generation termination
        gconfig.max_new_tokens += constants.pipe_parallel_world_size() - 1

        for mbid in range(num_micro_batches):
            tensor_buffer.put("kv_cache_reserved", mbid, False)
            tensor_buffer.put(
                "terminate",
                mbid,
                torch.tensor(0, dtype=torch.bool, device=self.module.device),
            )
            tensor_buffer.put("generated_idx", mbid, 0)
            batch_length = tensor_buffer.get("batch_lengths", mbid)
            tensor_buffer.put(
                "unfinished_sequences",
                mbid,
                torch.ones(
                    batch_length, dtype=torch.long, device=self.module.device
                ),
            )
            tensor_buffer.put("gen_token_ph", mbid, [])
            tensor_buffer.put("gen_logprob_ph", mbid, [])
            tensor_buffer.put("gen_logits_mask_ph", mbid, [])
            tensor_buffer.put("first_token", mbid, True)
            tensor_buffer.put("tokenizer", mbid, tokenizer)
            tensor_buffer.put("gconfig", mbid, gconfig)

        sched = schedule.GenerateSchedule(
            micro_batches=num_micro_batches,
            stages=constants.pipe_parallel_world_size(),
            stage_id=constants.pipe_parallel_rank(),
            max_new_tokens=gconfig.max_new_tokens,
        )

        def terminate_condition():
            return all(
                [
                    tensor_buffer.get("terminate", mbid)
                    for mbid in range(num_micro_batches)
                ]
            )

        _exec_pipe_schedule(
            self.module,
            tensor_buffer,
            instr_map=PipeGenInstrSet.INSTRUCTION_MAP,
            pipe_schedule=sched,
            terminate_condition=terminate_condition,
        )

        if not constants.is_last_pipe_stage():
            return None

        # Gather generation outputs, including generated tokens, logprobs, and logits_mask.
        generate_output = []
        for mbid in range(num_micro_batches):
            generate_output += [
                _gather_gen_output_from_list(
                    gen_token_ph=tensor_buffer.get(
                        "gen_token_ph", mbid, remove=True
                    ),
                    gen_logprob_ph=tensor_buffer.get(
                        "gen_logprob_ph", mbid, remove=True
                    ),
                    gen_logits_mask_ph=tensor_buffer.get(
                        "gen_logits_mask_ph", mbid, remove=True
                    ),
                )
            ]

        gen_tokens, log_probs, logits_mask = _gather_minibatch_gen_outputs(
            *list(zip(*generate_output))
        )

        return gen_tokens, log_probs, logits_mask, None, None

    def train_batch(
        self,
        engine: Any,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        loss_fn: Callable,
        version_steps: int,
        input_lens_for_partition: Optional[torch.Tensor] = None,
        num_micro_batches: Optional[int] = None,
        **loss_fn_kwargs,
    ):
        if not torch._C.is_grad_enabled():
            raise RuntimeError(
                f"train_batch() requires gradients enabled. Use eval_batch() instead."
            )

        if num_micro_batches is None:
            num_micro_batches = self.default_train_mbs

        tensor_buffer = TensorBuffer()
        for i in range(num_micro_batches):
            tensor_buffer.put("num_micro_batches", i, num_micro_batches)
            tensor_buffer.put("version_steps", i, version_steps)
            tensor_buffer.put("loss_fn", i, loss_fn)

        _prepare_input(
            module=self.module,
            tensor_buffer=tensor_buffer,
            num_micro_batches=num_micro_batches,
            seqlens_cpu=seqlens_cpu,
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
            store_kv_cache=False,
            loss_fn=loss_fn,
            input_lens_for_partition=input_lens_for_partition,
            **loss_fn_kwargs,
        )

        if isinstance(engine, DeepSpeedEngine):
            instr_map = PipeTrainInstrSet(
                backend="deepspeed",
                ds_engine=engine,
                num_micro_batches=num_micro_batches,
            ).INSTRUCTION_MAP
        elif isinstance(engine, MegatronEngine):
            instr_map = PipeTrainInstrSet(
                backend="megatron",
                megatron_engine=engine,
                num_micro_batches=num_micro_batches,
            ).INSTRUCTION_MAP
        else:
            raise NotImplementedError(
                f"Unknown backend type for training: {type(engine)}"
            )

        sched = schedule.TrainSchedule(
            micro_batches=num_micro_batches,
            stages=constants.pipe_parallel_world_size(),
            stage_id=constants.pipe_parallel_rank(),
        )
        _exec_pipe_schedule(
            module=self.module,
            tensor_buffer=tensor_buffer,
            instr_map=instr_map,
            pipe_schedule=sched,
        )

        agg_stats = None
        if constants.is_last_pipe_stage():
            stats = []
            for mbid in range(num_micro_batches):
                stats.append(tensor_buffer.get("stats", mbid))
            agg_stats = dict()
            for key in stats[0].keys():
                agg_stats[key] = torch.stack(
                    [stat[key] for stat in stats]
                ).sum()

        return agg_stats

    @torch.no_grad()
    def eval_batch(
        self,
        seqlens_cpu: List[int],
        packed_input_ids: torch.Tensor,
        cu_seqlens: torch.Tensor,
        loss_fn: Callable,
        input_lens_for_partition: Optional[torch.Tensor] = None,
        num_micro_batches: Optional[int] = None,
        **loss_fn_kwargs,
    ):
        if num_micro_batches is None:
            num_micro_batches = self.default_train_mbs

        tensor_buffer = TensorBuffer()
        for i in range(num_micro_batches):
            tensor_buffer.put("num_micro_batches", i, num_micro_batches)
            tensor_buffer.put("loss_fn", i, loss_fn)

        _prepare_input(
            module=self.module,
            tensor_buffer=tensor_buffer,
            num_micro_batches=num_micro_batches,
            seqlens_cpu=seqlens_cpu,
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
            store_kv_cache=False,
            loss_fn=loss_fn,
            input_lens_for_partition=input_lens_for_partition,
            **loss_fn_kwargs,
        )

        sched = schedule.InferenceSchedule(
            micro_batches=num_micro_batches,
            stages=constants.pipe_parallel_world_size(),
            stage_id=constants.pipe_parallel_rank(),
        )
        _exec_pipe_schedule(
            module=self.module,
            tensor_buffer=tensor_buffer,
            instr_map=PipeTrainForwardCommInstrSet.INSTRUCTION_MAP,
            pipe_schedule=sched,
        )

        agg_stats = None
        if constants.is_last_pipe_stage():
            stats = []
            for mbid in range(num_micro_batches):
                stats.append(tensor_buffer.get("stats", mbid))
            agg_stats = dict()
            for key in stats[0].keys():
                agg_stats[key] = torch.stack(
                    [stat[key] for stat in stats]
                ).sum()
        return agg_stats



from typing import Dict, Optional
import dataclasses
import os

import colorama
import deepspeed
import torch
import torch.distributed as dist
import tqdm

from realhf.base import constants
from realhf.base.namedarray import from_dict, NamedArray, recursive_apply
from realhf.impl.model.nn.real_llm_api import ReaLModel
import realhf.api.core.model_api as model_api
import realhf.base.logging as logging

logger = logging.getLogger("Packed Reward Modeling Interface", "benchmark")


def _paired_rw_loss_from_model_outputs(
    scores: torch.FloatTensor,
    packed_input_ids: torch.LongTensor,
    cu_seqlens: torch.IntTensor,
    group_factor: torch.FloatTensor,
    **kwargs,
):
    scores = scores[cu_seqlens[1:] - 1].view(-1, 2).float()
    loss = -(
        torch.nn.functional.logsigmoid(scores[:, 0] - scores[:, 1])
        * group_factor
    ).sum()

    # Logging.
    correct_predictions = (
        (scores[:, 0] > scores[:, 1]).count_nonzero().detach().float()
    )
    total_predictions = torch.tensor(
        scores.shape[0], dtype=torch.float32, device=scores.device
    )
    dist.all_reduce(
        correct_predictions,
        op=dist.ReduceOp.SUM,
        group=constants.data_parallel_group(),
    )
    dist.all_reduce(
        total_predictions,
        op=dist.ReduceOp.SUM,
        group=constants.data_parallel_group(),
    )
    pos_score_sum = scores[:, 0].sum().detach()
    max_pos_score = scores[:, 0].max(dim=0).values
    neg_score_sum = scores[:, 1].sum().detach()
    min_neg_score = scores[:, 1].min(dim=0).values
    dist.all_reduce(
        pos_score_sum,
        op=dist.ReduceOp.SUM,
        group=constants.data_parallel_group(),
    )
    dist.all_reduce(
        neg_score_sum,
        op=dist.ReduceOp.SUM,
        group=constants.data_parallel_group(),
    )
    loss_logging = loss.detach()
    dist.all_reduce(
        loss_logging,
        op=dist.ReduceOp.SUM,
        group=constants.data_parallel_group(),
    )
    dist.all_reduce(
        max_pos_score,
        op=dist.ReduceOp.MAX,
        group=constants.data_parallel_group(),
    )
    dist.all_reduce(
        min_neg_score,
        op=dist.ReduceOp.MIN,
        group=constants.data_parallel_group(),
    )

    return loss, dict(
        loss=loss_logging,
        correct_predictions=correct_predictions,
        total_predictions=total_predictions,
        pos_score=pos_score_sum,
        neg_score=neg_score_sum,
        max_pos_score=max_pos_score,
        min_neg_score=min_neg_score,
    )


@dataclasses.dataclass
class PairedRewardInterface(model_api.ModelInterface):
    enable_save: bool = True

    output_scaling: float = 1.0
    output_bias: float = 0.0

    # training log
    train_total_predictions: int = 0
    train_total_correct_predictions: int = 0

    @torch.no_grad()
    def inference(self, model: model_api.Model, data: NamedArray) -> NamedArray:
        data = recursive_apply(data, lambda x: x.to(model.device))
        packed_input_ids: torch.Tensor = data["packed_input_ids"]
        seqlens_cpu = data.metadata["seqlens"]
        max_seqlen = max(seqlens_cpu)
        cu_seqlens = torch.nn.functional.pad(
            torch.tensor(
                seqlens_cpu, dtype=torch.int32, device=model.device
            ).cumsum(0),
            (1, 0),
        )

        module: deepspeed.DeepSpeedEngine = model.module

        module.eval()

        r = module.forward(
            seqlens_cpu=data.metadata["seqlens"],
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
        )
        if r is None:
            return
        scores = r.float()

        scores = scores.squeeze(-1)[cu_seqlens[1:] - 1].float()  # [bs]
        scores = (scores - self.output_bias) * self.output_scaling

        ###################### logging ######################
        # input_ids = [packed_input_ids[start:end] for start, end in zip(cu_seqlens[:-1], cu_seqlens[1:])]
        # seq_strs = model.tokenizer.batch_decode(input_ids,
        #                                         clean_up_tokenization_spaces=False,
        #                                         skip_special_tokens=True)
        # for seq_str, score in zip(seq_strs, scores):
        #     logger.info(
        #         f"reward is {colorama.Fore.RED}{score.item()}{colorama.Style.RESET_ALL}, "
        #         f"sequence is: {colorama.Fore.YELLOW + colorama.Style.DIM}{seq_str}{colorama.Style.RESET_ALL}"
        #     )
        #####################################################

        res = from_dict(dict(scores=scores))
        res.register_metadata(**data.metadata)
        return res

    def train_step(
        self, model: model_api.Model, data: NamedArray
    ) -> NamedArray:
        data = recursive_apply(data, lambda x: x.to(model.device))

        packed_input_ids: torch.Tensor = data["packed_input_ids"]
        pair_lens = torch.tensor(
            data.metadata["seqlens"], dtype=torch.int32, device=model.device
        )
        neg_input_lens = pair_lens - data["pos_input_lens"]
        input_lens: torch.Tensor = torch.stack(
            [data["pos_input_lens"], neg_input_lens], 1
        ).view(-1)
        group_factor: torch.Tensor = data["group_factor"]
        cu_seqlens = torch.cat(
            [input_lens.new_zeros(1), input_lens.cumsum(0)], 0
        ).int()
        max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))

        module = model.module
        module.train()

        loss_fn_kwargs = dict(
            input_lens=pair_lens,
            group_factor=data["group_factor"],
        )
        stats = module.train_batch(
            seqlens_cpu=data.metadata["seqlens"],
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
            loss_fn=_paired_rw_loss_from_model_outputs,
            input_lens_for_partition=pair_lens,
            version_steps=model.version.global_step,
            **loss_fn_kwargs,
        )

        res = {}
        if stats:
            if constants.pipe_parallel_world_size() > 1:
                stats["max_pos_score"] /= (
                    constants.pipe_parallel_world_size() * 2
                )
                stats["min_neg_score"] /= (
                    constants.pipe_parallel_world_size() * 2
                )
            self.train_total_predictions += int(stats["total_predictions"])
            self.train_total_correct_predictions += int(
                stats["correct_predictions"]
            )
            res = dict(
                loss=float(stats["loss"] / stats["total_predictions"]),
                epoch_acc=self.train_total_correct_predictions
                / self.train_total_predictions,
                batch_acc=float(
                    stats["correct_predictions"] / stats["total_predictions"]
                ),
                avg_pos_score=float(
                    stats["pos_score"] / stats["total_predictions"]
                ),
                avg_neg_score=float(
                    stats["neg_score"] / stats["total_predictions"]
                ),
                total_predictions=int(stats["total_predictions"]),
                correct_predictions=int(stats["correct_predictions"]),
                max_pos_score=float(stats["max_pos_score"]),
                min_neg_score=float(stats["min_neg_score"]),
            )

        cur_epoch = model.version.epoch
        model.inc_version()
        if model.version.epoch > cur_epoch:
            self.train_total_predictions = (
                self.train_total_correct_predictions
            ) = 0

        return res

    def save(self, model: model_api.Model, save_dir: str):
        module = model.module
        if not isinstance(module, ReaLModel):
            module = module.module
        module.save_to_hf(
            tokenizer=model.tokenizer,
            save_dir=save_dir,
        )

    @torch.no_grad()
    def evaluate(
        self,
        model_: model_api.Model,
        eval_dataloader: torch.utils.data.DataLoader,
    ) -> Dict:
        device = model_.device
        model = model_.module

        model.eval()
        total_predictions = correct_predictions = 0
        losses = 0
        pos_score = neg_score = 0
        max_pos_score = -float("inf")
        min_neg_score = float("inf")

        for step, data in enumerate(tqdm.tqdm(eval_dataloader)):
            pair_lens = torch.tensor(
                data.metadata["seqlens"], dtype=torch.int32, device=model.device
            )
            data = recursive_apply(data, lambda x: x.to(device))

            packed_input_ids: torch.Tensor = data["packed_input_ids"]
            neg_input_lens = pair_lens - data["pos_input_lens"]
            assert (neg_input_lens > 0).all()
            input_lens = torch.stack(
                [data["pos_input_lens"], neg_input_lens], 1
            ).view(-1)
            group_factor: torch.Tensor = data["group_factor"]
            cu_seqlens = torch.cat(
                [input_lens.new_zeros(1), input_lens.cumsum(0)], 0
            ).int()
            max_seqlen = int(max(cu_seqlens[1:] - cu_seqlens[:-1]))

            loss_fn_kwargs = dict(
                input_lens=pair_lens,
                group_factor=data["group_factor"],
            )
            stats = model.eval_batch(
                seqlens_cpu=data.metadata["seqlens"],
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                loss_fn=_paired_rw_loss_from_model_outputs,
                input_lens_for_partition=pair_lens,
                **loss_fn_kwargs,
            )

            if stats:
                assert input_lens.shape[0] % 2 == 0
                losses += stats["loss"].item()
                correct_predictions += stats["correct_predictions"].item()
                total_predictions += stats["total_predictions"].item()
                pos_score += stats["pos_score"].item()
                neg_score += stats["neg_score"].item()
                max_pos_score = max(
                    max_pos_score, stats["max_pos_score"].item()
                )
                min_neg_score = min(
                    min_neg_score, stats["min_neg_score"].item()
                )

        if total_predictions > 0:
            return dict(
                loss=float(losses / total_predictions),
                acc=correct_predictions / total_predictions,
                pos_score=float(pos_score / total_predictions),
                neg_score=float(neg_score / total_predictions),
                correct_predictions=int(correct_predictions),
                total_predictions=int(total_predictions),
                max_pos_score=float(max_pos_score),
                min_neg_score=float(min_neg_score),
            )
        return dict()


model_api.register_interface("paired_rw", PairedRewardInterface)



import functools
import os
import re

import torch

from realhf.api.core.model_api import HF_MODEL_FAMILY_REGISTRY
from realhf.base.importing import import_module
from realhf.impl.model.conversion.hf_registry import HFModelRegistry
from realhf.impl.model.nn.real_llm_api import ReaLModel

# Import all HuggingFace model implementations.
import realhf.api.from_hf

# Import all model implementations.
_p = re.compile(r"^(?!.*__init__).*\.py$")
_filepath = os.path.dirname(__file__)
import_module(os.path.join(_filepath, "backend"), _p)
import_module(os.path.join(_filepath, "interface"), _p)
import_module(os.path.join(_filepath, "nn"), _p)

# Set PyTorch JIT options, following Megatron-LM.
if torch.cuda.is_available():
    torch._C._jit_set_profiling_executor(True)
    torch._C._jit_set_profiling_mode(True)
    torch._C._jit_override_can_fuse_on_cpu(False)
    torch._C._jit_override_can_fuse_on_gpu(False)
    torch._C._jit_set_texpr_fuser_enabled(False)
    torch._C._jit_set_nvfuser_enabled(True)
    torch._C._debug_set_autodiff_subgraph_inlining(False)

# Add HuggingFace hooks to ReaLModel.
_HF_REGISTRIES = {}


def _load_from_hf(
    model: ReaLModel, registry_name, load_dir: str, init_critic_from_actor: bool
):
    r = _HF_REGISTRIES[registry_name]
    setattr(
        model,
        "save_to_hf",
        functools.partial(_save_to_hf, model, registry_name),
    )
    return r.load(model, load_dir, init_critic_from_actor)


def _save_to_hf(model: ReaLModel, registry_name, tokenizer, save_dir: str):
    r = _HF_REGISTRIES[registry_name]
    r.save(model, tokenizer, save_dir)


def _config_from_hf(
    registry_name, hf_config=None, model_path=None, is_critic=False
):
    r = _HF_REGISTRIES[registry_name]
    return r.config_from_hf(hf_config, model_path, is_critic)


def _config_to_hf(registry_name, config):
    r = _HF_REGISTRIES[registry_name]
    return r.config_to_hf(config)


for name, helpers in HF_MODEL_FAMILY_REGISTRY.items():
    _HF_REGISTRIES[name] = r = HFModelRegistry(**helpers)

    _load_from_hf_ = functools.partialmethod(_load_from_hf, name)
    setattr(ReaLModel, f"from_{name}", _load_from_hf_)

    _save_to_hf_ = functools.partialmethod(_save_to_hf, name)
    setattr(ReaLModel, f"to_{name}", _save_to_hf_)

    _config_from_hf_ = functools.partial(_config_from_hf, name)
    setattr(ReaLModel, f"config_from_{name}", staticmethod(_config_from_hf_))

    _config_to_hf_ = functools.partial(_config_to_hf, name)
    setattr(ReaLModel, f"config_to_{name}", staticmethod(_config_to_hf_))



from typing import *
import dataclasses

import numpy as np
import torch

from realhf.api.core import model_api
from realhf.api.core.config import ModelName
from realhf.base import logging

try:
    import realhf._C.interval_op_cuda as interval_op_cuda
except ImportError:
    print(
        "interval_op_cuda not found. "
        "This should only appear on workers without CUDA supports."
    )

from .real_llm_base import (
    OutputHead,
    real_model_embed_param_count,
    real_model_embedding_param_keys,
    real_model_head_param_count,
    real_model_head_param_keys,
    real_model_tblock_param_count,
    real_model_tblock_param_keys,
    ReaLModelBlock,
    SequenceParallelActorHead,
    SequenceParallelCriticHead,
    VocabPositionEmbedding,
)
from .real_llm_parallel import (
    get_real_model_param_shape,
    intervals_partition_fn,
    mp_partition_key,
    partition_pipeline_layers,
    shape_partition_fn,
)

logger = logging.getLogger("FlattenParam")

MAX_PYTORCH_N_INTERVALS = 1024
CUDA_INTERVAL_OP_CHUNK_SIZE = 2048
_FLAT_PARAM_INDICES_CACHE = {}


@dataclasses.dataclass
class ContiguousParamSpec:
    start_idx: int
    end_idx: int
    shape: torch.Size


def _is_integer_list_contiguous(l: List[int]) -> bool:
    return np.all(np.array(l) == np.arange(len(l)) + l[0])


def _are_intervals_contiguous(l: List[Tuple[int, int]]) -> bool:
    l = sorted(l, key=lambda x: x[0])
    res = True
    for i in range(len(l) - 1):
        res &= l[i][1] == l[i + 1][0]
    return res


def recursive_getattr(obj, attr_string):
    attrs = attr_string.split(".")
    for attr in attrs:
        obj = getattr(obj, attr)
    return obj


def slice_intervals(
    tensor: torch.Tensor,
    intervals: torch.IntTensor,
    intervals_cpu: List[Tuple[int, int]],
    max_interval_size: int,
    output_size: int,
) -> torch.Tensor:
    assert len(tensor.shape) == 1
    if len(intervals_cpu) == 1:
        return tensor[intervals_cpu[0][0] : intervals_cpu[0][1]]
    elif len(intervals_cpu) <= MAX_PYTORCH_N_INTERVALS:
        return torch.cat([tensor[start:end] for start, end in intervals_cpu])

    interval_sizes = intervals[:, 1] - intervals[:, 0]
    offsets = torch.nn.functional.pad(
        interval_sizes.cumsum(0)[:-1], (1, 0), value=0
    )
    assert tensor.dtype == torch.half
    return interval_op_cuda.slice_intervals_cuda_half(
        tensor,
        intervals,
        interval_sizes,
        offsets,
        max_interval_size,
        output_size,
    )


def set_intervals(
    src: torch.Tensor,
    dst: torch.Tensor,
    intervals: torch.IntTensor,
    intervals_cpu: List[Tuple[int, int]],
    max_interval_size: int,
):
    assert len(dst.shape) == len(src.shape) == 1
    if len(intervals_cpu) <= MAX_PYTORCH_N_INTERVALS:
        offset = 0
        for i, j in intervals_cpu:
            dst[i:j] = src[offset : offset + j - i]
            offset += j - i
        assert offset == src.shape[0]
        return
    interval_sizes = intervals[:, 1] - intervals[:, 0]
    offsets = torch.nn.functional.pad(
        interval_sizes.cumsum(0)[:-1], (1, 0), value=0
    )
    interval_op_cuda.set_intervals_cuda_half(
        src,
        dst,
        intervals,
        interval_sizes,
        offsets,
        max_interval_size,
    )
    return


def build_param_spec(
    layer_indices: List[int],
    config: model_api.ReaLModelConfig,
    mp_size: int,
    sequence_parallel: bool,
) -> Tuple[Dict[str, ContiguousParamSpec], int]:
    if len(layer_indices) == 0:
        return {}, 0

    sd_keys = []
    for layer_idx in layer_indices:
        if layer_idx == 0:
            sd_keys += real_model_embedding_param_keys(config)
        elif layer_idx == config.n_layers + 1:
            sd_keys += real_model_head_param_keys(config)
        else:
            sd_keys += real_model_tblock_param_keys(config, layer_idx - 1)

    # In the reverse order as backpropagation, consistent with Megatron.
    sd_keys = reversed(sd_keys)

    param_spec = {}
    param_size = 0
    for k in sd_keys:
        shape = get_real_model_param_shape(
            k, config, mp_size, sequence_parallel
        )
        param_spec[k] = ContiguousParamSpec(
            param_size, param_size + int(np.prod(shape)), shape
        )
        param_size += int(np.prod(shape))
    return param_spec, param_size


def param_intervals_from_keys(
    model_name: ModelName,
    config: model_api.ReaLModelConfig,
    param_spec: Dict[str, ContiguousParamSpec],
    mp_size: int,
    sequence_parallel: bool,
    sd_keys: List[str],
    portion_size: int,
    portion_rank: int,
) -> List[int]:
    if portion_size == 1:
        start, end = None, None
        for k in sd_keys:
            if start is None or param_spec[k].start_idx < start:
                start = param_spec[k].start_idx
            if end is None or param_spec[k].end_idx > end:
                end = param_spec[k].end_idx
        return [(start, end)]

    intervals = []
    for k in sd_keys:
        if (
            model_name,
            k.split(".", 1)[1],
            mp_size,
            portion_rank,
            portion_size,
        ) not in _FLAT_PARAM_INDICES_CACHE:
            zero_start_intervals = mp_partition_key(
                k,
                get_real_model_param_shape(
                    k, config, mp_size, sequence_parallel
                ),
                portion_rank,
                portion_size,
                config,
                partition_fn=intervals_partition_fn,
            )
            _FLAT_PARAM_INDICES_CACHE[
                (
                    model_name,
                    k.split(".", 1)[1],
                    mp_size,
                    portion_rank,
                    portion_size,
                )
            ] = zero_start_intervals
        else:
            zero_start_intervals = _FLAT_PARAM_INDICES_CACHE[
                (
                    model_name,
                    k.split(".", 1)[1],
                    mp_size,
                    portion_rank,
                    portion_size,
                )
            ]
        intervals += (zero_start_intervals + param_spec[k].start_idx).tolist()
    # assert len(set([x[0] for x in intervals])) == len(intervals)
    intervals = sorted(intervals, key=lambda x: x[0])
    return intervals


def map_param_to_contigous_memory(
    layers: torch.nn.ModuleList,
    param_spec: Dict[str, ContiguousParamSpec],
    contiguous_param: torch.Tensor,
    layer_idx_offset: int,
):
    for local_layer_idx, l in enumerate(layers):
        layer_idx = local_layer_idx + layer_idx_offset
        for k, v in l.named_parameters():
            spec = param_spec[f"{layer_idx}.{k}"]
            old_param_data = v.data
            recursive_getattr(l, k).data = contiguous_param[
                spec.start_idx : spec.end_idx
            ].view(spec.shape)
            # This is for reward model. We should initialize the reward head instead of letting it be all-zero.
            if old_param_data.shape == spec.shape:
                v.data.copy_(old_param_data)
            else:
                assert old_param_data.shape == torch.Size([0]), (
                    old_param_data.shape,
                    spec.shape,
                )



import random

import numpy as np
import torch
import transformers


def set_random_seed(seed):
    transformers.set_seed(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)



from typing import Callable, Dict, List, Optional
import json

import numpy as np
import torch
import torch.utils.data

from realhf.api.core import data_api
from realhf.base import logging, namedarray

logger = logging.getLogger("Prompt Answer Dataset")


class PromptAnswerDataset(torch.utils.data.Dataset):

    def __init__(
        self,
        util: data_api.DatasetUtility,
        max_length: int,
        dataset_path: Optional[str] = None,
        dataset_builder: Optional[Callable[[], List[Dict]]] = None,
    ):
        """A dataset with prompts and corresponding answers. Usually used for SFT.

        Args:
            util (api.data.DatasetUtility): .
            max_length (Optional[int], optional): The maximum length of each sequence in the batch.
            dataset_path (Optional[str], optional): Path to the dataset json/jsonl file.
                The json/jsonl file should be a list of dictionary. Each element in the list should have
                a key "prompt" and a key "answer". Defaults to None.
            dataset_builder (Optional[Callable[[], List[Dict]]], optional): Alternative to dataset_path.
                A callable that returns a list of dictionary. Defaults to None.
        """
        self.util = util
        tokenizer = self.util.tokenizer

        data = data_api.load_shuffle_split_dataset(
            util, dataset_path, dataset_builder
        )

        seqs = [x["prompt"] + x["answer"] + tokenizer.eos_token for x in data]
        prompts = [x["prompt"] for x in data]

        self.tokens = tokenizer(
            seqs,
            padding=False,
            truncation=True,
            max_length=max_length,
            return_length=True,
            return_attention_mask=False,
        )
        prompt_tokens = tokenizer(
            prompts,
            padding=False,
            truncation=True,
            return_length=True,
            max_length=max_length,
            return_attention_mask=False,
        )

        prompt_lengths = prompt_tokens["length"]
        seq_lengths = self.tokens["length"]
        prompt_masks = []
        for i in range(len(self)):
            prompt_len = prompt_lengths[i]
            seqlen = self.tokens["length"][i]
            # seq = self.tokens["input_ids"][i]
            # prompt = prompt_tokens["input_ids"][i]
            # assert seq[:prompt_len] == prompt, (seq, prompt, prompt_len, seqlen)
            assert seqlen >= prompt_len, (seqlen, prompt_len)
            prompt_mask = [1] * prompt_len + [0] * (seqlen - prompt_len)
            prompt_masks.append(prompt_mask)

        self.prompt_masks = prompt_masks

        logger.info(
            f"Loaded Prompt Answer Dataset with INFO: "
            f"#seqs={len(self)}, "
            f"truncation length={max_length}, "
            f"avg prompt length={np.mean(prompt_lengths):.1f}, "
            f"avg answer length={np.mean(seq_lengths) - np.mean(prompt_lengths):.1f}",
        )

    def __len__(self):
        return len(self.tokens["input_ids"])

    def __getitem__(self, idx):
        d = {
            "packed_input_ids": torch.tensor(
                self.tokens["input_ids"][idx], dtype=torch.long
            ),
            "prompt_mask": torch.tensor(
                self.prompt_masks[idx], dtype=torch.bool
            ),
        }
        assert len(d["packed_input_ids"]) == len(d["prompt_mask"])
        seqlen = len(d["packed_input_ids"])
        x = namedarray.from_dict(d)
        x.register_metadata(seqlens=[seqlen])
        return x


data_api.register_dataset("prompt_answer", PromptAnswerDataset)



import os
import re

from realhf.base.importing import import_module

# Import all dataset implementations.
_p = re.compile(r"^(?!.*__init__).*\.py$")
_filepath = os.path.dirname(__file__)
import_module(_filepath, _p)



# Modified from https://github.com/vllm-project/vllm/blob/main/setup.py
from pathlib import Path
from typing import List, Set
import contextlib
import io
import os
import re
import subprocess
import warnings

from packaging.version import parse, Version
from torch.utils.cpp_extension import (
    BuildExtension,
    CUDA_HOME,
    CUDAExtension,
    ROCM_HOME,
)
import setuptools
import torch
import torch.utils.cpp_extension as torch_cpp_ext

ROOT_DIR = os.path.dirname(__file__)

# Supported NVIDIA GPU architectures.
NVIDIA_SUPPORTED_ARCHS = {"7.0", "7.5", "8.0", "8.6", "8.9", "9.0"}


def _is_cuda() -> bool:
    return os.getenv("REAL_CUDA", "0") == "1"


# Compiler flags.
CXX_FLAGS = ["-g", "-O3", "-std=c++17"]
NVCC_FLAGS = ["-O3", "-std=c++17"]

if _is_cuda() and CUDA_HOME is None:
    raise RuntimeError(
        "Cannot find CUDA_HOME. In GPU environment, CUDA must be available to build the package."
    )

ABI = 1 if torch._C._GLIBCXX_USE_CXX11_ABI else 0
CXX_FLAGS += [f"-D_GLIBCXX_USE_CXX11_ABI={ABI}"]
NVCC_FLAGS += [f"-D_GLIBCXX_USE_CXX11_ABI={ABI}"]


def glob(pattern: str):
    root = Path(__name__).parent
    return [str(p) for p in root.glob(pattern)]


def get_pybind11_include_path() -> str:
    pybind11_meta = subprocess.check_output(
        "python3 -m pip show pybind11", shell=True
    ).decode("ascii")
    for line in pybind11_meta.split("\n"):
        line = line.strip()
        if line.startswith("Location: "):
            return os.path.join(line.split(": ")[1], "pybind11", "include")


def get_nvcc_cuda_version(cuda_dir: str) -> Version:
    """Get the CUDA version from nvcc.

    Adapted from https://github.com/NVIDIA/apex/blob/8b7a1ff183741dd8f9b87e7bafd04cfde99cea28/setup.py
    """
    nvcc_output = subprocess.check_output(
        [cuda_dir + "/bin/nvcc", "-V"], universal_newlines=True
    )
    output = nvcc_output.split()
    release_idx = output.index("release") + 1
    nvcc_cuda_version = parse(output[release_idx].split(",")[0])
    return nvcc_cuda_version


def get_torch_arch_list() -> Set[str]:
    # TORCH_CUDA_ARCH_LIST can have one or more architectures,
    # e.g."8.0" or "7.5,8.0,8.6+PTX".Here, the "8.6+PTX" option asks the
    # compiler to additionally include PTX code that can be runtime - compiled
    # and executed on the 8.6 or newer architectures.While the PTX code will
    # not give the best performance on the newer architectures, it provides
    # forward compatibility.
    env_arch_list = os.environ.get("TORCH_CUDA_ARCH_LIST", None)
    if env_arch_list is None:
        return set()

    # List are separated by; or space.
    torch_arch_list = set(env_arch_list.replace(" ", ";").split(";"))
    if not torch_arch_list:
        return set()

    # Filter out the invalid architectures and print a warning.
    valid_archs = NVIDIA_SUPPORTED_ARCHS.union(
        {s + "+PTX" for s in NVIDIA_SUPPORTED_ARCHS}
    )
    arch_list = torch_arch_list.intersection(valid_archs)
    # If none of the specified architectures are valid, raise an error.
    if not arch_list:
        raise RuntimeError(
            "None of the CUDA architectures in `TORCH_CUDA_ARCH_LIST` env "
            f"variable ({env_arch_list}) is supported. "
            f"Supported CUDA architectures are: {valid_archs}."
        )
    invalid_arch_list = torch_arch_list - valid_archs
    if invalid_arch_list:
        warnings.warn(
            f"Unsupported CUDA architectures ({invalid_arch_list}) are "
            "excluded from the `TORCH_CUDA_ARCH_LIST` env variable "
            f"({env_arch_list}). Supported CUDA architectures are: "
            f"{valid_archs}.",
            stacklevel=2,
        )
    return arch_list


# First, check the TORCH_CUDA_ARCH_LIST environment variable.
compute_capabilities = get_torch_arch_list()

if _is_cuda() and not compute_capabilities:
    # If TORCH_CUDA_ARCH_LIST is not defined or empty, target all available
    # GPUs on the current machine.
    device_count = torch.cuda.device_count()
    for i in range(device_count):
        major, minor = torch.cuda.get_device_capability(i)
        if major < 7:
            raise RuntimeError(
                "GPUs with compute capability below 7.0 are not supported."
            )
        compute_capabilities.add(f"{major}.{minor}")

ext_modules = []

if _is_cuda():
    nvcc_cuda_version = get_nvcc_cuda_version(CUDA_HOME)
    if not compute_capabilities:
        # If no GPU is specified nor available, add all supported architectures
        # based on the NVCC CUDA version.
        compute_capabilities = NVIDIA_SUPPORTED_ARCHS.copy()
        if nvcc_cuda_version < Version("11.1"):
            compute_capabilities.remove("8.6")
        if nvcc_cuda_version < Version("11.8"):
            compute_capabilities.remove("8.9")
            compute_capabilities.remove("9.0")
    # Validate the NVCC CUDA version.
    if nvcc_cuda_version < Version("11.0"):
        raise RuntimeError(
            "CUDA 11.0 or higher is required to build the package."
        )
    if nvcc_cuda_version < Version("11.1") and any(
        cc.startswith("8.6") for cc in compute_capabilities
    ):
        raise RuntimeError(
            "CUDA 11.1 or higher is required for compute capability 8.6."
        )
    if nvcc_cuda_version < Version("11.8"):
        if any(cc.startswith("8.9") for cc in compute_capabilities):
            # CUDA 11.8 is required to generate the code targeting compute capability 8.9.
            # However, GPUs with compute capability 8.9 can also run the code generated by
            # the previous versions of CUDA 11 and targeting compute capability 8.0.
            # Therefore, if CUDA 11.8 is not available, we target compute capability 8.0
            # instead of 8.9.
            warnings.warn(
                "CUDA 11.8 or higher is required for compute capability 8.9. "
                "Targeting compute capability 8.0 instead.",
                stacklevel=2,
            )
            compute_capabilities = set(
                cc for cc in compute_capabilities if not cc.startswith("8.9")
            )
            compute_capabilities.add("8.0+PTX")
        if any(cc.startswith("9.0") for cc in compute_capabilities):
            raise RuntimeError(
                "CUDA 11.8 or higher is required for compute capability 9.0."
            )

    NVCC_FLAGS_PUNICA = NVCC_FLAGS.copy()

    # Add target compute capabilities to NVCC flags.
    for capability in compute_capabilities:
        num = capability[0] + capability[2]
        NVCC_FLAGS += ["-gencode", f"arch=compute_{num},code=sm_{num}"]
        if capability.endswith("+PTX"):
            NVCC_FLAGS += ["-gencode", f"arch=compute_{num},code=compute_{num}"]
        if int(capability[0]) >= 8:
            NVCC_FLAGS_PUNICA += [
                "-gencode",
                f"arch=compute_{num},code=sm_{num}",
            ]
            if capability.endswith("+PTX"):
                NVCC_FLAGS_PUNICA += [
                    "-gencode",
                    f"arch=compute_{num},code=compute_{num}",
                ]

    # Use NVCC threads to parallelize the build.
    if nvcc_cuda_version >= Version("11.2"):
        nvcc_threads = int(os.getenv("NVCC_THREADS", 8))
        num_threads = min(os.cpu_count(), nvcc_threads)
        NVCC_FLAGS += ["--threads", str(num_threads)]

    if nvcc_cuda_version >= Version("11.8"):
        NVCC_FLAGS += ["-DENABLE_FP8_E5M2"]

    # changes for punica kernels
    NVCC_FLAGS += torch_cpp_ext.COMMON_NVCC_FLAGS
    REMOVE_NVCC_FLAGS = [
        "-D__CUDA_NO_HALF_OPERATORS__",
        "-D__CUDA_NO_HALF_CONVERSIONS__",
        "-D__CUDA_NO_BFLOAT16_CONVERSIONS__",
        "-D__CUDA_NO_HALF2_OPERATORS__",
    ]
    for flag in REMOVE_NVCC_FLAGS:
        with contextlib.suppress(ValueError):
            torch_cpp_ext.COMMON_NVCC_FLAGS.remove(flag)

os.makedirs(os.path.join(ROOT_DIR, "realhf", "_C"), exist_ok=True)
if _is_cuda():
    cr_extension = CUDAExtension(
        name="realhf._C.custom_all_reduce",
        sources=[
            "csrc/custom_all_reduce/custom_all_reduce.cu",
            "csrc/custom_all_reduce/pybind.cpp",
        ],
        extra_compile_args={
            "cxx": CXX_FLAGS,
            "nvcc": NVCC_FLAGS,
        },
        libraries=["cuda"] if _is_cuda() else [],
    )
    # ext_modules.append(cr_extension)

    interval_extension = CUDAExtension(
        name="realhf._C.interval_op_cuda",
        sources=[
            "csrc/interval_op/interval_ops.cu",
        ],
        extra_compile_args={
            "cxx": CXX_FLAGS,
            "nvcc": NVCC_FLAGS,
        },
        libraries=["cuda"] if _is_cuda() else [],
    )
    ext_modules.append(interval_extension)

    gae_extension = CUDAExtension(
        name="realhf._C.cugae",
        sources=[
            "csrc/cugae/gae.cu",
        ],
        extra_compile_args={
            "cxx": CXX_FLAGS,
            "nvcc": NVCC_FLAGS
            + [
                "--expt-relaxed-constexpr",
                "--expt-extended-lambda",
                "--use_fast_math",
            ],
        },
        libraries=["cuda"] if _is_cuda() else [],
    )
    ext_modules.append(gae_extension)

search_extension = setuptools.Extension(
    name="realhf._C.mdm_search",
    sources=[
        "csrc/search/search.cpp",
        "csrc/search/rpc.cpp",
        "csrc/search/device_mesh.cpp",
        "csrc/search/simulate.cpp",
    ],
    language="c++",
    extra_compile_args=[
        "-O3",
        "-Wall",
        "-shared",
        "-std=c++11",
        "-fPIC",
        "-std=c++17",
    ],
    include_dirs=[
        os.path.join(os.path.dirname(__file__), "csrc", "search"),
        get_pybind11_include_path(),
    ],
)
ext_modules.append(search_extension)

if os.getenv("REAL_NO_EXT", "0") == "1":
    ext_modules = []

setuptools.setup(
    name="realhf",
    ext_modules=ext_modules,
    cmdclass={"build_ext": BuildExtension},
    packages=setuptools.find_packages(),
)



# Implements a simple name resolving service, which can be considered as a distributed key-value dict.
from typing import Callable, List, Optional
import dataclasses
import getpass
import os
import queue
import random
import shutil
import socket
import threading
import time
import uuid

from realhf.base import logging, security, timeutil
from realhf.base.cluster import spec as cluster_spec

logger = logging.getLogger("name-resolve")


class ArgumentError(Exception):
    pass


class NameEntryExistsError(Exception):
    pass


class NameEntryNotFoundError(Exception):
    pass


class NameRecordRepository:

    def __del__(self):
        try:
            self.reset()
        except Exception as e:
            logger.info(f"Exception ignore when deleting NameResolveRepo {e}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reset()

    def add(
        self,
        name,
        value,
        delete_on_exit=True,
        keepalive_ttl=None,
        replace=False,
    ):
        """Creates a name record in the central repository.

        In our semantics, the name record repository is essentially a multimap (i.e. Dict[str, Set[str]]).
        This class keeps a single name->value map, where the name can be non-unique while the value has to be.
        The class also deletes the (name, value) pair on exits (__exit__/__del__) if opted-in. In case of
        preventing unexpected exits (i.e. process crashes without calling graceful exits), an user may also
        want to specify time_to_live and call touch() regularly to allow a more consistent

        Args:
            name: The key of the record. It has to be a valid path-like string; e.g. "a/b/c". If the name
                already exists, the behaviour is defined by the `replace` argument.
            value: The value of the record. This can be any valid string.
            delete_on_exit: If the record shall be deleted when the repository closes.
            keepalive_ttl: If not None, adds a time-to-live in seconds for the record. The repository
                shall keep pinging the backend service with at least this frequency to make sure the name
                entry is alive during the lifetime of the repository. On the other hand, specifying this
                prevents stale keys caused by the scenario that a Python process accidentally crashes before
                calling delete().
            replace: If the name already exists, then replaces the current value with the supplied value if
                `replace` is True, or raises exception if `replace` is False.
        """
        raise NotImplementedError()

    def add_subentry(self, name, value, **kwargs):
        """Adds a sub-entry to the key-root `name`. The values is retrievable by get_subtree() given that no
        other entries use the name prefix.
        """
        sub_name = name.rstrip("/") + "/" + str(uuid.uuid4())[:8]
        self.add(sub_name, value, **kwargs)
        return sub_name

    def delete(self, name):
        """Deletes an existing record."""
        raise NotImplementedError()

    def clear_subtree(self, name_root):
        """Deletes all records whose names start with the path root name_root; specifically, whose name either
        is `name_root`, or starts with `name_root.rstrip("/") + "/"`.
        """
        raise NotImplementedError()

    def get(self, name):
        """Returns the value of the key. Raises NameEntryNotFoundError if not found."""
        raise NotImplementedError()

    def get_subtree(self, name_root):
        """Returns all values whose names start with the path root name_root; specifically, whose name either
        is `name_root`, or starts with `name_root.rstrip("/") + "/"`.
        """
        raise NotImplementedError()

    def find_subtree(self, name_root):
        """Returns all KEYS whose names start with the path root name_root."""
        raise NotImplementedError()

    def wait(self, name, timeout=None, poll_frequency=1):
        """Waits until a name appears.

        Raises:
             TimeoutError: if timeout exceeds.
        """
        start = time.monotonic()
        while True:
            try:
                return self.get(name)
            except NameEntryNotFoundError:
                pass
            if timeout is None or timeout > 0:
                time.sleep(
                    poll_frequency + random.random() * 0.1
                )  # To reduce concurrency.
            if timeout is not None and time.monotonic() - start > timeout:
                raise TimeoutError(
                    f"Timeout waiting for key '{name}' ({self.__class__.__name__})"
                )

    def reset(self):
        """Deletes all entries added via this repository instance's add(delete_on_exit=True)."""
        raise NotImplementedError()

    def watch_names(
        self,
        names: List,
        call_back: Callable,
        poll_frequency=15,
        wait_timeout=300,
    ):
        """Watch a name, execute call_back when key is deleted."""
        if isinstance(names, str):
            names = [names]

        q = queue.Queue(maxsize=len(names))
        for _ in range(len(names) - 1):
            q.put(0)

        def wrap_call_back():
            try:
                q.get_nowait()
            except queue.Empty:
                logger.info(
                    f"Key {names} is gone. Executing callback {call_back}"
                )
                call_back()

        for name in names:
            t = threading.Thread(
                target=self._watch_thread_run,
                args=(name, wrap_call_back, poll_frequency, wait_timeout),
                daemon=True,
            )
            t.start()

    def _watch_thread_run(self, name, call_back, poll_frequency, wait_timeout):
        self.wait(name, timeout=wait_timeout, poll_frequency=poll_frequency)
        while True:
            try:
                self.get(name)
                time.sleep(poll_frequency + random.random())
            except NameEntryNotFoundError:
                call_back()
                break


class MemoryNameRecordRepository(NameRecordRepository):
    """Stores all the records in a thread-local memory. Note that this is most likely for testing purposes:
    any distributed application is impossible to use this.
    """

    def __init__(self, log_events=False):
        self.__store = {}
        self.__log_events = log_events

    def add(
        self,
        name,
        value,
        delete_on_exit=True,
        keepalive_ttl=None,
        replace=False,
    ):
        if self.__log_events:
            print(f"NameResolve: add {name} {value}")
        if name in self.__store and not replace:
            raise NameEntryExistsError(
                f"K={name} V={self.__store[name]} V2={value}"
            )
        assert isinstance(value, str)
        self.__store[name] = value

    def touch(self, name, value, new_time_to_live):
        raise NotImplementedError()

    def delete(self, name):
        if self.__log_events:
            print(f"NameResolve: delete {name}")
        if name not in self.__store:
            raise NameEntryNotFoundError(f"K={name}")
        del self.__store[name]

    def clear_subtree(self, name_root):
        if self.__log_events:
            print(f"NameResolve: clear_subtree {name_root}")
        name_root = name_root.rstrip("/")
        for name in list(self.__store):
            if (
                name_root == "/"
                or name == name_root
                or name.startswith(name_root + "/")
            ):
                del self.__store[name]

    def get(self, name):
        if name not in self.__store:
            raise NameEntryNotFoundError(f"K={name}")
        r = self.__store[name]
        if self.__log_events:
            print(f"NameResolve: get {name} -> {r}")
        return r

    def get_subtree(self, name_root):
        if self.__log_events:
            print(f"NameResolve: get_subtree {name_root}")
        name_root = name_root.rstrip("/")
        rs = []
        for name, value in self.__store.items():
            if (
                name_root == "/"
                or name == name_root
                or name.startswith(name_root + "/")
            ):
                rs.append(value)
        return rs

    def find_subtree(self, name_root):
        if self.__log_events:
            print(f"NameResolve: find_subtree {name_root}")
        rs = []
        for name, value in self.__store.items():
            if name.startswith(name_root):
                rs.append(name)
        rs.sort()
        return rs

    def reset(self):
        self.__store = {}


class NfsNameRecordRepository(NameRecordRepository):
    RECORD_ROOT = f"{cluster_spec.fileroot}/name_resolve/"

    def __init__(self, **kwargs):
        self.__to_delete = set()

    @staticmethod
    def __dir_path(name):
        return os.path.join(NfsNameRecordRepository.RECORD_ROOT, name)

    @staticmethod
    def __file_path(name):
        return os.path.join(NfsNameRecordRepository.__dir_path(name), "ENTRY")

    def add(
        self,
        name,
        value,
        delete_on_exit=True,
        keepalive_ttl=None,
        replace=False,
    ):
        path = self.__file_path(name)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        if os.path.isfile(path) and not replace:
            raise NameEntryExistsError(path)
        local_id = str(uuid.uuid4())[:8]
        with open(path + f".tmp.{local_id}", "w") as f:
            f.write(str(value))
        os.rename(path + f".tmp.{local_id}", path)
        if delete_on_exit:
            self.__to_delete.add(name)

    def delete(self, name):
        path = self.__file_path(name)
        if not os.path.isfile(path):
            raise NameEntryNotFoundError(path)
        os.remove(path)
        while True:
            path = os.path.dirname(path)
            if path == NfsNameRecordRepository.RECORD_ROOT:
                break
            if len(os.listdir(path)) > 0:
                break
            shutil.rmtree(path, ignore_errors=True)
        if name in self.__to_delete:
            self.__to_delete.remove(name)

    def clear_subtree(self, name_root):
        dir_path = self.__dir_path(name_root)
        if os.path.isdir(dir_path):
            logger.info("Removing name resolve path: %s", dir_path)
            shutil.rmtree(dir_path)
        else:
            logger.info("No such name resolve path: %s", dir_path)

    def get(self, name):
        path = self.__file_path(name)
        if not os.path.isfile(path):
            raise NameEntryNotFoundError(path)
        with open(path, "r") as f:
            return f.read().strip()

    def get_subtree(self, name_root):
        dir_path = self.__dir_path(name_root)
        rs = []
        if os.path.isdir(dir_path):
            for item in os.listdir(dir_path):
                try:
                    rs.append(self.get(os.path.join(name_root, item)))
                except NameEntryNotFoundError:
                    pass
        return rs

    def find_subtree(self, name_root):
        dir_path = self.__dir_path(name_root)
        rs = []
        if os.path.isdir(dir_path):
            for item in os.listdir(dir_path):
                rs.append(os.path.join(name_root, item))
        rs.sort()
        return rs

    def reset(self):
        for name in list(self.__to_delete):
            try:
                self.delete(name)
            except:
                pass
        self.__to_delete = {}


class RedisNameRecordRepository(NameRecordRepository):
    _IS_FRL = socket.gethostname().startswith("frl")
    REDIS_HOST = "redis" if _IS_FRL else "localhost"
    REDIS_PASSWORD = security.read_key("redis") if _IS_FRL else None
    REDIS_DB = 0
    KEEPALIVE_POLL_FREQUENCY = 1

    @dataclasses.dataclass
    class _Entry:
        value: str
        keepalive_ttl: Optional[int] = None
        keeper: Optional[timeutil.FrequencyControl] = None

    def __init__(self, **kwargs):
        from redis.backoff import ExponentialBackoff
        from redis.retry import Retry
        import redis

        super().__init__()
        self.__lock = threading.Lock()
        self.__redis = redis.Redis(
            host=RedisNameRecordRepository.REDIS_HOST,
            password=RedisNameRecordRepository.REDIS_PASSWORD,
            db=RedisNameRecordRepository.REDIS_DB,
            socket_timeout=60,
            retry_on_timeout=True,
            retry=Retry(ExponentialBackoff(180, 60), 3),
        )
        self.__entries = {}
        self.__keepalive_running = True
        self.__keepalive_thread = threading.Thread(
            target=self.__keepalive_thread_run, daemon=True
        )
        self.__keepalive_thread.start()

    def __del__(self):
        self.__keepalive_running = False
        self.__keepalive_thread.join(timeout=5)
        self.reset()
        self.__redis.close()

    def add(
        self, name, value, delete_on_exit=True, keepalive_ttl=10, replace=False
    ):
        # deprecated parameter: delete_on_exit, now every entry has a default keepalive_ttl=10 seconds
        if name.endswith("/"):
            raise ValueError(f"Entry name cannot end with '/': {name}")

        with self.__lock:
            keepalive_ttl = int(keepalive_ttl * 1000)
            assert (
                keepalive_ttl > 0
            ), f"keepalive_ttl in milliseconds must >0: {keepalive_ttl}"
            if (
                self.__redis.set(name, value, px=keepalive_ttl, nx=not replace)
                is None
            ):
                raise NameEntryExistsError(
                    f"Cannot set Redis key: K={name} V={value}"
                )

            # touch every 1/3 of keepalive_ttl to prevent Redis from deleting the key
            # after program exit, redis will automatically delete key in keepalive_ttl
            self.__entries[name] = self._Entry(
                value=value,
                keepalive_ttl=keepalive_ttl,
                keeper=timeutil.FrequencyControl(
                    frequency_seconds=keepalive_ttl / 1000 / 3
                ),
            )

    def delete(self, name):
        with self.__lock:
            self.__delete_locked(name)

    def __delete_locked(self, name):
        if name in self.__entries:
            del self.__entries[name]
        if self.__redis.delete(name) == 0:
            raise NameEntryNotFoundError(
                f"No such Redis entry to delete: {name}"
            )

    def clear_subtree(self, name_root):
        with self.__lock:
            count = 0
            for name in list(self.__find_subtree_locked(name_root)):
                try:
                    self.__delete_locked(name)
                    count += 1
                except NameEntryNotFoundError:
                    pass
            logger.info("Deleted %d Redis entries under %s", count, name_root)

    def get(self, name):
        with self.__lock:
            return self.__get_locked(name)

    def __get_locked(self, name):
        r = self.__redis.get(name)
        if r is None:
            raise NameEntryNotFoundError(f"No such Redis entry: {name}")
        return r.decode()

    def get_subtree(self, name_root):
        with self.__lock:
            rs = []
            for name in self.__find_subtree_locked(name_root):
                rs.append(self.__get_locked(name))
            rs.sort()
            return rs

    def find_subtree(self, name_root):
        with self.__lock:
            return list(sorted(self.__find_subtree_locked(name_root)))

    def reset(self):
        with self.__lock:
            count = 0
            for name in list(self.__entries):
                try:
                    self.__delete_locked(name)
                    count += 1
                except NameEntryNotFoundError:
                    pass
            self.__entries = {}
            logger.info("Reset %d saved Redis entries", count)

    def __keepalive_thread_run(self):
        while self.__keepalive_running:
            time.sleep(self.KEEPALIVE_POLL_FREQUENCY)
            with self.__lock:
                for name, entry in self.__entries.items():
                    if entry.keeper is not None and entry.keeper.check():
                        r = self.__redis.set(
                            name, entry.value, px=entry.keepalive_ttl
                        )
                        if r is None:
                            logger.error(
                                "Failed touching Redis key: K=%s V=%s",
                                name,
                                entry.value,
                            )

    def __find_subtree_locked(self, name_root):
        pattern = name_root + "*"
        return [k.decode() for k in self.__redis.keys(pattern=pattern)]

    def _testonly_drop_cached_entry(self, name):
        """Used by unittest only to simulate the case that the Python process crashes and the key is
        automatically removed after TTL."""
        with self.__lock:
            del self.__entries[name]
            print("Testonly: dropped key:", name)


def make_repository(type_="nfs", **kwargs):
    if type_ == "memory":
        return MemoryNameRecordRepository(**kwargs)
    elif type_ == "nfs":
        return NfsNameRecordRepository(**kwargs)
    elif type_ == "redis":
        return RedisNameRecordRepository(**kwargs)
    else:
        raise NotImplementedError(f"No such name resolver: {type_}")


# DEFAULT_REPOSITORY_TYPE = "redis" if socket.gethostname().startswith("frl") else "nfs"
DEFAULT_REPOSITORY_TYPE = "nfs"
DEFAULT_REPOSITORY = make_repository(DEFAULT_REPOSITORY_TYPE)
add = DEFAULT_REPOSITORY.add
add_subentry = DEFAULT_REPOSITORY.add_subentry
delete = DEFAULT_REPOSITORY.delete
clear_subtree = DEFAULT_REPOSITORY.clear_subtree
get = DEFAULT_REPOSITORY.get
get_subtree = DEFAULT_REPOSITORY.get_subtree
find_subtree = DEFAULT_REPOSITORY.find_subtree
wait = DEFAULT_REPOSITORY.wait
reset = DEFAULT_REPOSITORY.reset
watch_names = DEFAULT_REPOSITORY.watch_names


def reconfigure(*args, **kwargs):
    global DEFAULT_REPOSITORY, DEFAULT_REPOSITORY_TYPE
    global add, add_subentry, delete, clear_subtree, get, get_subtree, find_subtree, wait, reset, watch_names
    DEFAULT_REPOSITORY = make_repository(*args, **kwargs)
    DEFAULT_REPOSITORY_TYPE = args[0]
    add = DEFAULT_REPOSITORY.add
    add_subentry = DEFAULT_REPOSITORY.add_subentry
    delete = DEFAULT_REPOSITORY.delete
    clear_subtree = DEFAULT_REPOSITORY.clear_subtree
    get = DEFAULT_REPOSITORY.get
    get_subtree = DEFAULT_REPOSITORY.get_subtree
    find_subtree = DEFAULT_REPOSITORY.find_subtree
    wait = DEFAULT_REPOSITORY.wait
    reset = DEFAULT_REPOSITORY.reset
    watch_names = DEFAULT_REPOSITORY.watch_names



# Copied from the vLLM project: https://github.com/vllm-project/vllm
from contextlib import contextmanager
from typing import Optional

import torch
import torch.distributed as dist

import realhf.base.constants as constants
import realhf.base.logging as logging

try:
    from custom_all_reduce import custom_ar
    import pynvml
except ImportError:
    # For AMD GPUs
    custom_ar = None
    pynvml = None

logger = logging.getLogger(__name__)

_CA_HANDLE = None
_IS_CAPTURING = False
_SUPPORTED_WORLD_SIZES = [2, 4, 6, 8]


def init_custom_ar() -> None:
    global _CA_HANDLE
    if _CA_HANDLE is not None:
        return
    rank = constants.model_parallel_rank()
    world_size = constants.model_parallel_world_size()
    if world_size == 1:
        # No need to initialize custom allreduce for single GPU case.
        return

    if world_size not in _SUPPORTED_WORLD_SIZES:
        logger.warn(
            "Custom allreduce is disabled due to an unsupported world size: "
            "%d. Supported world sizes: %s. To silence this warning, specify"
            "disable_custom_all_reduce=True explicitly.",
            world_size,
            str(_SUPPORTED_WORLD_SIZES),
        )
        return
    if not _can_p2p(rank, world_size):
        logger.warn(
            "Custom allreduce is disabled because your platform lacks GPU P2P"
            " capability. To silence this warning, specify"
            "disable_custom_all_reduce=True explicitly."
        )
        return
    _CA_HANDLE = CustomAllreduce(rank, world_size)


def begin_capture() -> None:
    global _IS_CAPTURING
    _IS_CAPTURING = True


def end_capture() -> None:
    global _IS_CAPTURING
    _IS_CAPTURING = False


def is_capturing() -> bool:
    return _IS_CAPTURING and _CA_HANDLE is not None


def get_handle() -> Optional["CustomAllreduce"]:
    return _CA_HANDLE


def is_initialized() -> bool:
    return _CA_HANDLE is not None


@contextmanager
def capture():
    try:
        begin_capture()
        yield
    finally:
        end_capture()
        handle = get_handle()
        if handle is not None:
            handle.register_graph_buffers()


def custom_all_reduce(input: torch.Tensor) -> Optional[torch.Tensor]:
    ca_handle = get_handle()
    # when custom allreduce is disabled, this will be None
    if ca_handle is None:
        return
    if is_capturing():
        if torch.cuda.is_current_stream_capturing():
            if ca_handle.should_custom_ar(input):
                return ca_handle.all_reduce_reg(input)
        else:
            if ca_handle.should_custom_ar(input):
                # if warm up, mimic the allocation pattern
                # since custom allreduce is out-of-place
                return torch.empty_like(input)
    else:
        # note: outside of cuda graph context,
        # custom allreduce incurs a cost of cudaMemcpy, which should
        # be small(<=1% of overall latency) compared to the performance
        # gains of using custom kernels
        if ca_handle.should_custom_ar(input):
            return ca_handle.all_reduce_unreg(input)


@contextmanager
def _nvml():
    try:
        pynvml.nvmlInit()
        yield
    finally:
        pynvml.nvmlShutdown()


# query if the set of gpus are fully connected by nvlink (1 hop)
@_nvml()
def _is_full_nvlink(rank, world_size):
    handle = pynvml.nvmlDeviceGetHandleByIndex(rank)
    for i in range(world_size):
        if i != rank:
            try:
                link_state = pynvml.nvmlDeviceGetNvLinkState(handle, i)
                if not link_state:
                    return False
            except pynvml.NVMLError as error:
                logger.info(
                    f'NVLink detection failed with message "{str(error)}". '
                    "This is normal if your machine has no NVLink equipped"
                )
                return False
    return True


def _can_p2p(rank: int, world_size: int) -> bool:
    for i in range(world_size):
        if i == rank:
            continue
        if not torch.cuda.can_device_access_peer(rank, i):
            return False
    return True


class CustomAllreduce:

    # max_size: max supported allreduce size
    def __init__(self, rank, world_size, max_size=8192 * 1024) -> None:
        # buffers memory are owned by this Python class and passed to C++
        # meta data composes of two parts: meta data for synchronization
        # (256 bytes) and a temporary buffer for storing intermediate
        # allreduce results.
        self.meta = torch.zeros(
            custom_ar.meta_size() + max_size, dtype=torch.uint8, device="cuda"
        )
        # This is a pre-registered IPC buffer. In eager mode, input tensors
        # are first copied into this buffer before allreduce is performed
        self.buffer = torch.empty(max_size, dtype=torch.uint8, device="cuda")
        # This is a buffer for storing the tuples of pointers pointing to
        # IPC buffers from all ranks. Each registered tuple has size of
        # 8*world_size bytes where world_size is at most 8. Allocating 8MB
        # is enough for 131072 such tuples. The largest model I've seen only
        # needs less than 10000 of registered tuples.
        self.rank_data = torch.empty(
            8 * 1024 * 1024, dtype=torch.uint8, device="cuda"
        )
        self.max_size = max_size
        self.world_size = world_size
        handles, offsets = self._get_ipc_meta(self.meta)
        self.full_nvlink = _is_full_nvlink(rank, world_size)
        self._ptr = custom_ar.init_custom_ar(
            self.meta, self.rank_data, handles, offsets, rank, self.full_nvlink
        )
        self.fast_cond = self.full_nvlink or world_size <= 2
        self.register_buffer(self.buffer)

    def _get_ipc_meta(self, inp: torch.Tensor):
        data = inp.untyped_storage()._share_cuda_()
        shard_data = (
            data[1],  # ipc handle to base ptr
            data[3],  # offset of base ptr
        )
        return self._gather_ipc_meta(shard_data)

    def _gather_ipc_meta(self, shard_data):
        all_data = [None] * self.world_size
        dist.all_gather_object(all_data, shard_data)

        handles = []
        offsets = []
        for i in range(len(all_data)):
            handles.append(all_data[i][0])
            offsets.append(all_data[i][1])
        return handles, offsets

    def register_buffer(self, inp: torch.Tensor):
        handles, offsets = self._get_ipc_meta(inp)
        custom_ar.register_buffer(self._ptr, inp, handles, offsets)

    def register_graph_buffers(self):
        handle, offset = custom_ar.get_graph_buffer_ipc_meta(self._ptr)
        handles, offsets = self._gather_ipc_meta((bytes(handle), offset))
        logger.info("Registering %d cuda graph addresses", len(offset))
        custom_ar.register_graph_buffers(self._ptr, handles, offsets)

    def should_custom_ar(self, inp: torch.Tensor):
        return custom_ar.should_custom_ar(
            inp, self.max_size, self.world_size, self.full_nvlink
        )

    # all reduce, assuming inp tensor is IPC registered with register_buffer,
    # or, in the context of cuda graphs, register_graph_buffers
    def all_reduce_reg(self, inp: torch.Tensor, out: torch.Tensor = None):
        if out is None:
            out = torch.empty_like(inp)
        custom_ar.all_reduce_reg(self._ptr, inp, out)
        return out

    # all reduce, assuming inp tensor is NOT IPC registered
    def all_reduce_unreg(self, inp: torch.Tensor, out: torch.Tensor = None):
        if out is None:
            out = torch.empty_like(inp)
        custom_ar.all_reduce_unreg(self._ptr, inp, self.buffer, out)
        return out

    def close(self):
        if self._ptr:
            custom_ar.dispose(self._ptr)
            self._ptr = 0

    def __del__(self):
        self.close()



from typing import List
import dataclasses

from realhf.api.core.dfg import MFCDef, ModelInterface, ModelInterfaceType
from realhf.api.core.system_api import *
from realhf.api.quickstart.dataset import PairedComparisonDatasetConfig
from realhf.api.quickstart.device_mesh import AllocationConfig
from realhf.api.quickstart.entrypoint import register_quickstart_exp
from realhf.api.quickstart.model import (
    get_real_model_config,
    ModelTrainEvalConfig,
    OptimizerConfig,
)
from realhf.experiments.common.common import CommonExperimentConfig


@dataclasses.dataclass
class RWConfig(CommonExperimentConfig):
    """Pairwise reward modeling experiment configuration.

    It is a subclass of :class:`CommonExperimentConfig`,
    so all CLI options in the base class are available.

    :param total_train_epochs: Total number of training epochs
        (i.e., the number of times the training dataset is iterated).
    :type total_train_epochs: int
    :param save_freq_steps: Save the model every this number of steps.
        "step" is an optimizer step or a single update of model parameters.
        If None, the model will not be saved during training.
        The directory to save the model will be automatically resolved
        and prompted in the terminal when the experiment starts.
    :type save_freq_steps: Optional[int]
    :param eval_freq_epochs: Evaluate the model every this number of epochs.
        If None, the model will not be evaluated during training.
    :type eval_freq_epochs: Optional[int]
    :param is_sft_lora: Whether LoRA was used for SFT.
        If so, the saved SFT model should only contain LoRA parameters.
        Since LoRA is currently not supported for SFT,
        this option is not used for now.
    :type is_sft_lora: bool
    :param sft_lora_path: Path to the LoRA model for SFT.
        Since LoRA is currently not supported for SFT,
        this option is not used for now.
    :param model: Model runtime configuration.
    :type model: ModelTrainEvalConfig
    :param allocation: Device allocation and parallelism configuration.
    :type allocation: AllocationConfig
    :param dataset: Dataset configuration.
    :type dataset: PairedComparisonDatasetConfig
    """

    total_train_epochs: int = 1
    save_freq_steps: Optional[int] = 20
    eval_freq_epochs: Optional[int] = 1
    is_sft_lora: bool = False
    sft_lora_path: Optional[str] = None
    model: ModelTrainEvalConfig = dataclasses.field(
        default_factory=ModelTrainEvalConfig
    )
    allocation: AllocationConfig = dataclasses.field(
        default_factory=AllocationConfig
    )

    dataset: PairedComparisonDatasetConfig = dataclasses.field(
        default_factory=PairedComparisonDatasetConfig
    )

    def __post_init__(self):
        assert (
            not self.is_sft_lora and self.sft_lora_path is None
        ), "LoRA is not supported for now."
        self.model.init_critic_from_actor = True

    @property
    def models(self):
        return {
            "default": self.model,
        }

    @property
    def rpcs(self):
        interface = ModelInterface("paired_rw")
        rpc = MFCDef(
            model_name=ModelName("default", 0),
            interface_type=ModelInterfaceType.TRAIN_STEP,
            interface_impl=interface,
            model_type=self.model.type,
            model_path=self.model.path,
            input_data=["packed_input_ids", "group_factor", "pos_input_lens"],
            log_return_value=True,
            min_n_seqs=self.dataset.train_bs_n_seqs,
            max_n_seqs=self.dataset.train_bs_n_seqs,
        )
        return {"default": rpc}

    @property
    def allocations(self):
        return {"default": self.allocation}

    @property
    def datasets(self):
        return [
            Dataset(
                "rw_pair",
                args=dict(
                    max_length=self.dataset.max_seqlen,
                    max_pairs_per_prompt=self.dataset.max_pairs_per_prompt,
                    dataset_path=self.dataset.train_path,
                ),
            )
        ]

    @property
    def eval_datasets(self):
        return [
            Dataset(
                "rw_pair",
                args=dict(
                    max_length=self.dataset.max_seqlen,
                    max_pairs_per_prompt=self.dataset.max_pairs_per_prompt,
                    dataset_path=self.dataset.valid_path,
                ),
            )
        ]

    @property
    def eval_dataloader(self):
        return DataLoader(
            "packed_eval", args=dict(batch_size=self.dataset.valid_bs_n_seqs)
        )

    @property
    def tokenizer_name_or_path(self):
        return self.model.path

    @property
    def exp_ctrl(self):
        return ExperimentSaveEvalControl(
            total_train_epochs=self.total_train_epochs,
            save_frequency_steps=self.save_freq_steps,
            eval_frequency_epochs=self.eval_freq_epochs,
        )


register_quickstart_exp("rw", RWConfig)
